"""Utility package, primarily focussed on data science work.

Works on standard import, although specific submodules may be imported directly

```python
# Import package and useful submodules for your project
import utc

data_path = utc.path_create("some", "path")
data = utc.loader.load_data(data_path)
```

```python
# Import specific functions you want
from utc.core import path_create  # Use * to get all
from utc import path_create  # Equivalent as core is imported by default

path_create("some", "path")
```

```python
# Import a submodule as its own module
import utc.loader as uloader

data = uloader.load_data(data_path)
```

"""
import os
import subprocess

__version__ = "0.13.2"  # In case something goes terribly wrong

# Fallback
if __version__ == "placeholder":
    # Try to get latest tag from master for versioning purposes in development builds
    try:
        latest_tag = (
            subprocess.check_output(
                ["git", "describe", "master", "--tags"],
                cwd=os.path.dirname(os.path.abspath(__file__)),
            )
            .decode()
            .strip()
        )
        try:
            # Only works for git >=2.22
            cur_branch = (
                subprocess.check_output(
                    ["git", "branch", "--show-current"],
                    cwd=os.path.dirname(os.path.abspath(__file__)),
                )
                .decode()
                .strip()
            )
        except:
            # Fallback, works for git >=1.63
            # Not perfect but should always return
            cur_branch = (
                subprocess.check_output(
                    ["git", "rev-parse", "--abbrev-ref", "HEAD"],
                    cwd=os.path.dirname(os.path.abspath(__file__)),
                )
                .decode()
                .strip()
            )

        latest_ver = latest_tag.split("-")[0]

        __version__ = latest_ver
        if cur_branch != "master":
            __version__ += "-{}".format(cur_branch)

    # Secondary fallback in case git isn't available
    except:
        from datetime import date

        __version__ = "dev_{}".format(date.today())


from utc.core import *

from utc import ml
from utc.ml import losses
from utc.ml import models

from utc import loader
from utc import plotting
from utc import windowing
from utc import data_manip
