"""Functions and attributes that are generally useful.
"""

import os
import sys
import copy
import types
import errno
import pickle
import random
import string
import smtplib
import contextlib

import pandas as pd
import numpy as np

from functools import partial
from collections.abc import Iterable

from typing import Dict


class LabGolem(object):  # pragma: no cover
    """Basic assistant for easy sending of emails to .shef.ac.uk accounts.

    By default makes use of the hidden IT Services mail server to send emails to any
    shef.ac.uk email account without the need for usernames/passwords.
    """

    def __init__(self, to_addr, send_addr="labgolem@gmail.com", server=None):
        """Setup LabGolem with a target address.

        * `to_addr`: address to send emails to
        * `send_addr`: address to from
        * `server`: `smtplib.SMTP(..)` server object, defaults to shef.ac.uk mailhost
        """
        self.to_addr = to_addr
        self.send_addr = send_addr

        if server is None:
            self.server = smtplib.SMTP("mailhost.shef.ac.uk", 587)
        else:
            self.server = server

    def email(self, subject="Test Mail", body=""):
        """Send an email to the previously set address

        * `subject`: subject line of email
        * `body`: Interior text of email
        """
        content = "\r\n".join(
            [
                "To: {}".format(self.to_addr),
                "Content-type: text/html",
                "From: {}".format(self.send_addr),
                "Subject: {}".format(subject),
                "",
                body,
            ]
        )

        self.server.sendmail(self.send_addr, self.to_addr, content)
        print("Email sent")


class SelfRefDict(dict):
    """Extension of a dictionary that can reference itself

    Use lambda functions in entries e.g. `"a": lambda self: self['b'] + 1`.

    Set up by passing a normal dictionary: `SelfRefDict({"x": x})`.
    """

    def __getitem__(self, key):
        item = dict.__getitem__(self, key)
        return item(self) if callable(item) else item

    def to_dict(self):
        new_dict = {}
        for key in self.keys():
            value = self[key]
            new_dict[key] = value

        return new_dict


@contextlib.contextmanager
def dummy_context_mgr():  # pragma: no cover
    """Dummy context manager for using with "with" statements.

    Use when you need a context manager for syntax reasons but don't want it to
    do anything.
    """
    yield None


def slugify(input_string, to_lower=True):
    """Sanitises a string by removing non-alpha characters and converting spaces
    to underscores.

    * `to_lower`: if `True` then convert input string to lower case.

    Particularly useful for filenames.
    """
    valid_chars = frozenset("-_.() {}{}".format(string.ascii_letters, string.digits))

    if to_lower:
        input_string = input_string.lower()

    return "".join(c for c in input_string if c in valid_chars).replace(" ", "_")


def get_nth_key(some_dict: Dict, n: int = 0) -> any:
    """Get `n`th key in a dictionary.

    Order of dictionaries is consistent with order of creation
    """
    from itertools import islice

    if n >= len(some_dict):
        raise IndexError("Dictionary index out of range")

    if n < 0:  # Wraparound to allow negative indexing
        n += len(some_dict)

    it = iter(some_dict)

    # Iterate to interesting point, suppressing StopIteration errors
    next(islice(it, n, n), None)

    return next(it)


def get_environment(
    super_globals,
    include_current_imports=False,
    return_dict=True,
    exact_spec=True,
    additional_dependencies=None,
    env_path=None,
    env_name=None,
    env_channels=None,
):  # pragma: no cover
    """Get user imported modules and their versions as ver strings.

    - `super_globals`: set to `globals()`; necessary to ensure we look at modules from
      the correct scope
    - `include_current_imports`: if True then include a "current_imports" section with details
      on what is currently imported and any renaming e.g. `import numpy as np`
    - `return_dict`: if True then return yaml dictionary
    - `exact_spec`: if True then use = for versions, otherwise use >=
    - `additional_dependencies`: list of extra dependencies to include, e.g. `pip`
    - `env_path`: optional path to an anaconda environment file in which to save version
      information
    - `env_name`: what to call the conda environment if saving to .yml
    - `env_channels`: list of channels to use for Anaconda, defaults to conda-forge and
      "defaults"

    Standard usage:
    ```python
    from utc import print_imports

    print_imports(globals(), env_path="environment.yml")
    ```

    Later the environment can be regenerated in `conda with the following:
    `conda env create -f environment.yml`

    """
    import yaml
    import platform

    ignore_list = ["__builtin__", "__builtins__", "__main__"] + list(
        sys.builtin_module_names
    )

    if running_in_jupyter():
        ignore_list += ["IPython"]  # Assume any normal import would be lower case
        ignore_list += ["ipywidgets"]  # Correct for *most* cases

    class MyDumper(yaml.Dumper):  # Shim for misappropriated yaml spec in conda
        def increase_indent(self, flow=False, indentless=False):
            return super(MyDumper, self).increase_indent(flow, False)

    if env_channels is None:
        env_channels = ["conda-forge", "defaults"]

    if additional_dependencies is None:
        additional_dependencies = []

    env_dict = {
        "channels": env_channels,
        "dependencies": additional_dependencies,
    }

    if env_name is None:
        env_dict["name"]: "env-name"
    else:
        env_dict["name"]: env_name

    spec_qualifier = "="
    if not exact_spec:
        spec_qualifier = ">="

    imported_list = []
    module_list = []
    for name, val in super_globals.items():
        if isinstance(val, types.ModuleType) and val.__name__ not in ignore_list:
            if hasattr(val, "__version__"):
                env_dict["dependencies"].append(
                    "{}{}{}".format(
                        val.__name__, spec_qualifier, val.__version__.split("-")[0]
                    )
                )

            if val.__name__ == name:
                imported_list.append("import {}".format(val.__name__))
            else:
                imported_list.append("import {} as {}".format(val.__name__, name))

            module_list.append(val.__name__)

        elif isinstance(val, types.FunctionType):
            module_top_level = val.__module__.split(".")[0]
            if module_top_level in ignore_list:
                continue

            if (
                module_top_level not in ignore_list
                and module_top_level not in module_list
            ):
                env_dict["dependencies"].append("{}".format(module_top_level))
                module_list.append(val.__name__)

            if val.__name__ == name:
                imported_list.append(
                    "from {} import {}".format(val.__module__, val.__name__)
                )
            else:
                imported_list.append(
                    "from {} import {} as {}".format(val.__module__, val.__name__, name)
                )

    env_dict["dependencies"] = list(set(env_dict["dependencies"]))
    env_dict["dependencies"].sort()
    env_dict["dependencies"] = [
        "python={}".format(platform.python_version())
    ] + env_dict["dependencies"]

    if include_current_imports:
        imported_list.sort(reverse=True)
        env_dict["current_imports"] = imported_list

    if env_path is not None:
        with open(env_path, "w") as file:
            yaml.dump(env_dict, file, Dumper=MyDumper, default_flow_style=False)

    if return_dict:
        return env_dict


class TerminalFormat:
    """Utility class for manipulating terminal formatting.

    Simply add `TerminalFormat.SOME_VAR` to your print statement. Formatting
    won't be removed until `TerminalFormat.END` is printed.
    """

    PURPLE = "\033[95m"
    CYAN = "\033[96m"
    DARKCYAN = "\033[36m"
    BLUE = "\033[94m"
    GREEN = "\033[92m"
    YELLOW = "\033[93m"
    RED = "\033[91m"
    GREY = "\033[30m"
    MAGENTA = "\033[35m"
    WHITE = "\033[37m"

    HIGHLIGHT_GREY = "\033[40m"
    HIGHLIGHT_RED = "\033[41m"
    HIGHLIGHT_GREEN = "\033[42m"
    HIGHLIGHT_YELLOW = "\033[43m"
    HIGHLIGHT_BLUE = "\033[44m"
    HIGHLIGHT_MAGENTA = "\033[45m"
    HIGHLIGHT_CYAN = "\033[46m"
    HIGHLIGHT_WHITE = "\033[47m"

    BOLD = "\033[1m"
    DARK = "\033[2m"
    UNDERLINE = "\033[4m"
    BLINK = "\033[5m"
    REVERSE = "\033[7m"
    CONCEALED = "\033[8m"

    END = "\033[0m"


def uprint(input_string, modifiers=None, **kwargs):  # pragma: no cover
    """Wrap print to simplify adding of TerminalFormat constants.

    * `modifiers`: list of strings that match TerminalFormat attribute names, case
        insensitive
    * Accepts any keyword arguments for the standard print function

    See TerminalFormat for list of possible values. Adds an end terminator after the
    string - note this will clear all formatting.
    """
    if modifiers is not None:
        if not isinstance(modifiers, list):
            modifiers = [modifiers]

        for mod in modifiers:
            input_string = eval("TerminalFormat.{}".format(mod.upper())) + str(
                input_string
            )

        input_string = input_string + TerminalFormat.END

    print(input_string, **kwargs)


def running_in_jupyter():  # pragma: no cover
    """Return True if running in a jupyter shell, False otherwise."""
    try:
        if "zmqshell" in str(type(get_ipython())):
            return True
    except:
        return False

    return False


def get_notebook_path():
    """Get the absolute path of the current Notebook or None if it cannot be determined

    Only works with security token or no password configurations
    """
    from notebook import notebookapp
    import urllib
    import json
    import ipykernel

    connection_file = os.path.basename(ipykernel.get_connection_file())
    kernel_id = connection_file.split("-", 1)[1].split(".")[0]

    for srv in notebookapp.list_running_servers():
        try:
            # No token and no password
            if srv["token"] == "" and not srv["password"]:
                req = urllib.request.urlopen(srv["url"] + "api/sessions")

            else:  # Otherwise hijack token
                req = urllib.request.urlopen(
                    srv["url"] + "api/sessions?token=" + srv["token"]
                )

            sessions = json.load(req)
            for sess in sessions:
                if sess["kernel"]["id"] == kernel_id:
                    return os.path.join(srv["notebook_dir"], sess["notebook"]["path"])
        except:
            pass  # Stale entries may cause errors as well as passwords etc

    return None


def z_score_lookup(n, confidence=0.95, sides=2):
    """Get the z score for a given a confidence level.

    Uses student t distribution for n <= 30, z otherwise. Uses central limit
    theorem i.e. passes n-1 to t distribution

    * `n`: degrees of freedom
    * `confidence`: confidence interval [0, 1]
    * `sides`: 1-sided or 2 sided test
    """
    import scipy.stats

    if confidence < 0 or confidence > 1:
        raise AttributeError("Confidence must be on range [0, 1]")

    if sides == 1:
        prob = confidence

    elif sides == 2:
        prob = (1 + confidence) / 2.0

    else:
        raise AttributeError("Sides must be 1 or 2")

    if n <= 30:
        conf_multiplier = scipy.stats.t.ppf(prob, n - 1)
    else:
        conf_multiplier = scipy.stats.norm.ppf(prob)

    return conf_multiplier


def reset_random_seeds(random_nb_seed=0, set_tf_seed=False):  # pragma: no cover
    """Reset Python core, numpy and, optionally, Tensorflow random seeds."""
    from random import seed as random_set_seed
    from numpy.random import seed as np_set_seed

    random_set_seed(random_nb_seed)
    np_set_seed(random_nb_seed)

    if set_tf_seed:
        import tensorflow as tf  # Full import necessary (partial fails)

        tf.random.set_seed(random_nb_seed)


def path_create(path, *paths):  # pragma: no cover
    """os.path.join except it creates folders if they don't exist."""
    new_path = os.path.join(path, *paths)

    try:
        os.makedirs(new_path)
    except OSError as e:
        if e.errno != errno.EEXIST:
            raise

    return new_path


def get_mask_from_indices(data, indices, level=None):
    """Get a mask for where indices are in a DataFrame, allows partial match.

    E.g. if you have a Multi-Index with 3 levels you can get a mask for all data
    that matches the top two levels.

    * `indices`: list of indices to try and find in index
    * `level`: what level to reset the index to for partial matches
      - If `None` then this will be calculated based on how long the target
        indices are versus the length of the actual index
    """
    if not isinstance(indices, list):
        indices = [indices]

    if level is None:
        if isinstance(indices[0], Iterable):
            index_levels = len(indices[0])
        else:
            index_levels = 1

        level = [x for x in range(-1, -(len(data.index[0]) - index_levels + 1), -1)]

    return data.reset_index(level=level).index.isin(indices)


def split_list_by_percent(some_list, p1, p2=None, shuffle=True, seed=None):
    """Split a list into 2 parts with given percentages

    * `some_list`: relevant list (won't be modified)
    * `p1`: percentage [0, 1] to include in first returned list
    * `p2`: if not `None`, percentage [0, 1] to include in second returned list
      - Automatically set to `1 - p1` if None
    * `shuffle`: if `True` then split randomly otherwise split into first and
      second half
    * `seed`: if not `None` then set the seed in `random` before shuffling the
      data
    """
    if seed is not None:
        random.seed(seed)

    if p2 is None:
        p2 = max([1 - p1, 0])

    if p1 + p2 != 1.0:
        raise RuntimeError("Split percentages (p1, p2) must add up to 1.")

    tmp_list = some_list
    if shuffle:
        tmp_list = copy.deepcopy(some_list)
        random.shuffle(tmp_list)

    nb_indices = len(tmp_list)
    nb_p1 = int(np.ceil(nb_indices * p1))

    return tmp_list[:nb_p1], tmp_list[nb_p1:]


def save_object(obj, filepath):  # pragma: no cover
    """Wrapper for simple saving with pickle."""
    with open(filepath, "wb") as out_file:
        pickle.dump(obj, out_file, pickle.HIGHEST_PROTOCOL)


def load_object(filepath):  # pragma: no cover
    """Wrapper for simple loading with pickle."""
    with open(filepath, "rb") as in_file:
        return pickle.load(in_file)


def get_col_names_with_dtype(df, dtypes, ignore_cols_with_nan=False):
    """Get names of all columns in a pandas dataframe with specified dtype."""
    if not isinstance(dtypes, list):
        dtypes = [dtypes]

    relevant_cols = []
    for col in df:
        if df[col].dtypes in dtypes:
            if ignore_cols_with_nan and df[col].isnull().values.any():
                continue

            relevant_cols.append(col)

    return relevant_cols


def get_git_origin():
    """Return current repo origin (or None on failure)"""
    from subprocess import check_output

    repo_origin = None
    try:
        repo_origin = (
            check_output(["git", "config", "--get", "remote.origin.url"])
            .decode("latin-1")
            .strip()
        )
    except:
        pass

    return repo_origin


def get_git_commit_hash():
    """Return current full commit hash (or None on failure)"""
    from subprocess import check_output

    repo_origin = None
    try:
        repo_origin = (
            check_output(["git", "rev-parse", "HEAD"]).decode("latin-1").strip()
        )
    except:
        pass

    return repo_origin


html_table_style = """
<style>
    table a:link {
        color: #666;
        font-weight: bold;
        text-decoration:none;
    }
    table a:visited {
        color: #999999;
        font-weight:bold;
        text-decoration:none;
    }
    table a:active,
    table a:hover {
        color: #bd5a35;
        text-decoration:underline;
    }
    table {
        font-family:Arial, Helvetica, sans-serif;
        color:#666;
        font-size:12px;
        text-shadow: 1px 1px 0px #fff;
        background:#eaebec;
        margin:20px;
        border:#ccc 1px solid;

        -moz-border-radius:3px;
        -webkit-border-radius:3px;
        border-radius:3px;

        -moz-box-shadow: 0 1px 2px #d1d1d1;
        -webkit-box-shadow: 0 1px 2px #d1d1d1;
        box-shadow: 0 1px 2px #d1d1d1;
    }
    table th {
        padding:21px 25px 22px 25px;
        border-top:1px solid #fafafa;
        border-bottom:1px solid #e0e0e0;

        background: #ededed;
        background: -webkit-gradient(linear, left top, left bottom, from(#ededed), to(#ebebeb));
        background: -moz-linear-gradient(top,  #ededed,  #ebebeb);
    }
    table th:first-child {
        text-align: left;
        padding-left:20px;
    }
    table tr:first-child th:first-child {
        -moz-border-radius-topleft:3px;
        -webkit-border-top-left-radius:3px;
        border-top-left-radius:3px;
    }
    table tr:first-child th:last-child {
        -moz-border-radius-topright:3px;
        -webkit-border-top-right-radius:3px;
        border-top-right-radius:3px;
    }
    table tr {
        text-align: center;
        padding-left:20px;
    }
    table td:first-child {
        text-align: left;
        padding-left:20px;
        border-left: 0;
    }
    table td {
        padding:18px;
        border-top: 1px solid #ffffff;
        border-bottom:1px solid #e0e0e0;
        border-left: 1px solid #e0e0e0;

        background: #fafafa;
        background: -webkit-gradient(linear, left top, left bottom, from(#fbfbfb), to(#fafafa));
        background: -moz-linear-gradient(top,  #fbfbfb,  #fafafa);
    }
    table tr.even td {
        background: #f6f6f6;
        background: -webkit-gradient(linear, left top, left bottom, from(#f8f8f8), to(#f6f6f6));
        background: -moz-linear-gradient(top,  #f8f8f8,  #f6f6f6);
    }
    table tr:last-child td {
        border-bottom:0;
    }
    table tr:last-child td:first-child {
        -moz-border-radius-bottomleft:3px;
        -webkit-border-bottom-left-radius:3px;
        border-bottom-left-radius:3px;
    }
    table tr:last-child td:last-child {
        -moz-border-radius-bottomright:3px;
        -webkit-border-bottom-right-radius:3px;
        border-bottom-right-radius:3px;
    }
    table tr:hover td {
        background: #f2f2f2;
        background: -webkit-gradient(linear, left top, left bottom, from(#f2f2f2), to(#f0f0f0));
        background: -moz-linear-gradient(top,  #f2f2f2,  #f0f0f0);
    }
</style>
"""
"""String of a generic 'style' element for nicely styling html tables.

Built in [Firefox's Stylus extension](https://addons.mozilla.org/en-US/firefox/addon/styl-us/).

"""

pretty_print_df = partial(
    pd.option_context, "display.precision", 3, "max_rows", 30, "max_columns", 20
)
"""Use to make nice printouts of pandas DataFrames

Use as a context i.e. with a `with` statement
```python
with pretty_print_df():
    df
```
"""
