import copy
import numpy as np


def add_step_anomaly(
    data, start_position=0, duration=1, abs_power=0, rel_power=0, inplace=False
):
    """Add a step anomaly to some signal

    * `data`: array-like
    * `start_position`: the position in the data where the anomaly should begin
        (absolute)
    * `duration`: how long the anomaly should last (in steps)
    * `abs_power`: the absolute magnitude of the anomaly, additive with
        `rel_power`
    * `rel_power`: the magnitude of the anomaly relative to the rest of the data
        (multiplier, additive with `abs_power`)
    * `inplace`: if true then attempt to manipulate in place

    Final value calculated as `data[i] += abs_power + rel_power*data_range`.

    Use negative values to create a step down and positive values for a step up
    """
    if not inplace:
        data = copy.deepcopy(data)

    data_range = np.max(data) - np.min(data)

    for i in range(duration):
        data[start_position + i] += abs_power + rel_power * data_range

    if not inplace:
        return data


def add_guassian_noise(
    data,
    target_snr_db=None,
    target_noise_db=None,
    abs_sigma=None,
    fs=None,
    f_cutoff=None,
    filter_order=None,
    inplace=False,
):
    """Add guassian noise to a signal (or signals)

    By default white guassian noise is applied. If `fs`, `f_cutoff`, and
    `filter_order` are not None

    * `data`: array-like
    * `target_snr`: generate noise to create a desired signal-noise ratio (db)
    * `target_db`: generate noise for a specific power (db)
    * `abs_sigma`: generate noise for a specific standard deviation
    * `fs`: *optional*, if defined then apply a low-pass butterworth filter to
      the noise signal prior to adding to the data based on this sampling
      frequency.
    * `f_cutoff`: *optional*, if defined then apply a low-pass butterworth
      filter to the noise signal prior to adding to the data with this cut-off
      frequency (Hz)
    * `filter_order`: *optional*,if `f_cutoff` is set then use this order for the
      butterworth filter
    * `inplace`: if true then attempt to manipulate in place
    """
    # Validate inputs
    count = 0
    if target_snr_db is not None:
        count += 1

    if target_noise_db is not None:
        count += 1

    if abs_sigma is not None:
        count += 1

    if count != 1:
        raise AttributeError("Exactly one noise power argument must be set.")

    if not inplace:
        data = copy.deepcopy(data)

    # Determine noise amplitude
    if target_snr_db is not None:
        signal_power = data ** 2
        sig_power_mean = np.mean(signal_power)
        sig_power_mean_db = 10 * np.log10(sig_power_mean)

        noise_power_mean_db = sig_power_mean_db - target_snr_db
        target_noise_scale = np.sqrt(10 ** (noise_power_mean_db / 10))

    if target_noise_db is not None:
        target_noise_scale = np.sqrt(10 ** (target_noise_db / 10))

    if abs_sigma is not None:
        target_noise_scale = abs_sigma

    # Make noise
    noise = np.random.normal(0, target_noise_scale, data.shape)

    # Apply low pass filter to noise if desired
    if fs is not None and f_cutoff is not None and filter_order is not None:
        from scipy import signal

        # Filter design
        b, a = signal.butter(filter_order, f_cutoff, btype="lowpass", fs=fs)

        # Apply to noise
        if len(noise.shape) == 2:
            for i in range(noise.shape[1]):
                noise[:, i] = signal.filtfilt(b, a, noise[:, i], padtype=None)

        else:
            noise = signal.filtfilt(b, a, noise, padtype=None)

    data += noise

    if not inplace:
        return data


def map_all_bad_to_missing(df, bad_values=None, inplace=False, augment_col=None):
    """Map values to NaN to indicate missing.

    * `bad_values`: list of values to consider bad, defaults to `[-1, -55555, -99999]`
    * `inplace`: whether to do in place
    * `augment_col`: optional name of a column to append that indicates which
      rows have been manipulated
    """
    if bad_values is None:
        bad_values = [-1, -55555, -99999]

    mask = df.isin(bad_values)

    if inplace:
        cur_data = df

    else:
        cur_data = copy.deepcopy(df)

    cur_data[mask] = np.NaN

    if augment_col is not None:
        new_mask = cur_data.isin([np.NaN])  # Pick up any existing nans
        cur_data[augment_col] = new_mask.any(axis=1)

    if inplace:
        return None

    else:
        return cur_data


def map_to_boolean(series):
    """Maps strings/numbers to actual booleans.

    The strings "True", "true", "False" and "false" will be mapped to their
    boolean counterparts. Zeros will be mapped to `false`, other numbers
    """
    map_dict = {
        "true": True,
        "false": False,
        "True": True,
        "False": False,
        True: True,
        False: False,
    }

    return series.map(map_dict)


def get_all_missing_mask(df):
    """Get a mask of which rows have at least 1 NaN in a dataframe

    Wraps pandas functionality.
    """
    return df.isnull().to_numpy().any(axis=1)


def series_to_categorical_df(series):  # pragma: no cover
    """Convert a categorical column into N boolean columns, return new dataframe.

    Wrapper around panda's `str.get_dummies`, columns will be called `X_is_Y`.
    Where X is the original column name and Y is the category. N is the number
    of different categories.
    """
    new_df = series.str.get_dummies().astype(bool)
    new_df_cols = ["{}_is_{}".format(series.name, col) for col in new_df.columns]
    new_df.columns = new_df_cols

    return new_df


def get_cruise_mask(
    altitude_data, minimum_cruise_altitude, nb_buffer=25, min_duration=50, max_delta=100
):  # pragma: no cover
    """Get a mask of where cruise happens in a dataset.

    * `nb_buffer`: number of points to ignore at start and end of cruise to
      catch edge effects

    **Experimental function, always check results.**
    """
    data_mask = np.zeros((altitude_data.shape[0],), dtype=bool)

    cruise_start_idx = get_first_level_off_point(
        altitude_data, minimum_cruise_altitude, min_duration, max_delta
    )

    cruise_start_idx_reverse = get_first_level_off_point(
        altitude_data[::-1], minimum_cruise_altitude, min_duration, max_delta
    )

    # Check sensible
    if cruise_start_idx is None or cruise_start_idx_reverse is None:
        print("Warning: cruise not found.")
        return data_mask

    cruise_start_idx += nb_buffer
    cruise_end_idx = altitude_data.shape[0] - (cruise_start_idx_reverse + nb_buffer) + 1

    if cruise_end_idx - cruise_start_idx <= 0:
        print("Warning: final cruise too short, try reducing buffer.")
        return data_mask

    data_mask[cruise_start_idx:cruise_end_idx] = True
    return data_mask


def get_first_level_off_point(
    data, min_value, min_duration=50, max_delta=100
):  # pragma: no cover
    """Get where data first reaches a steady state.

    **Experimental function, always check results.**
    """
    for win_end in range(min_duration, data.shape[0] + 1):
        win_start = win_end - min_duration

        cur_window = data[win_start:win_end]

        # Not yet at minimum value
        if np.any(cur_window < min_value):
            continue

        # If too large range of values
        if np.max(cur_window) - np.min(cur_window) > max_delta:
            continue

        return win_start

    return None


def clip_all_to_steady_state_range(
    data, ss_col, max_delta, valid_range=None, longest_only=False
):  # pragma: no cover
    """Clip data to steady state.

    * `data`: data - intended to be a single file but could be several
    * `ss_col`: the data column to decide steady state upon
    * `valid_range`: `[min, max]`: values outside this range will be completely
      ignored
    * `max_delta`: if the data changes by more than this value between
      timepoints it is considered to not be steady state
    * `longest_only`: whether to only return the longest continuous segment

    **Experimental function, always check results.**
    """

    # Catch case where range is the wrong way around, raise since indicative of other
    # issues
    if valid_range is not None and valid_range[0] > valid_range[1]:
        raise ValueError(
            "Range should take form [min, max], values: [{} {}]".format(
                valid_range[0], valid_range[1]
            )
        )

    clipped_data = copy.deepcopy(data)

    rapid_change_mask = np.full((data.shape[0],), True)
    for i in range(data.shape[0] - 1):
        if np.abs(data[ss_col].iloc[i] - data[ss_col].iloc[i + 1]) < max_delta:
            rapid_change_mask[i] = False

    clipped_data.loc[rapid_change_mask] = np.NaN

    if valid_range is not None:
        below_thres_mask = clipped_data[ss_col] < valid_range[0]
        above_thres_mask = clipped_data[ss_col] > valid_range[1]
        clipped_data.loc[below_thres_mask] = np.NaN
        clipped_data.loc[above_thres_mask] = np.NaN

    if longest_only:
        nan_mask = np.concatenate(
            ([True], np.isnan(clipped_data[ss_col].values), [True])
        )

        # Find edges in mask and pair out
        edge_idxs = np.flatnonzero(nan_mask[1:] != nan_mask[:-1]).reshape(-1, 2)

        if edge_idxs.shape[0] > 0:
            # Get max interval limits
            start_idx, end_idx = edge_idxs[(edge_idxs[:, 1] - edge_idxs[:, 0]).argmax()]

            # New mask
            longest_mask = np.full((data.shape[0],), True)
            longest_mask[start_idx:end_idx] = False
            clipped_data.loc[longest_mask] = np.NaN

    return clipped_data
