import os
import math
import copy
import errno
import pickle
import random
import numpy as np
import pandas as pd

from ..core import running_in_jupyter, html_table_style, get_col_names_with_dtype

from scipy.stats import spearmanr, pearsonr


def _make_cache_params(cache_folder, **kwargs):
    """Utility to ensure cache paths and parameter details are always the same.
    """
    param_cache_path = os.path.join(cache_folder, ".param_cache")
    data_cache_path = os.path.join(cache_folder, ".data_cache")

    # Use as many of the below as are available
    kwarg_banlist = [
        "data",
        "load_from_cache",
        "save_to_cache",
        "override_cache",
        "cache_dir",
        "cache_folder",
    ]

    param_str = ""
    for key in kwargs:
        if key not in kwarg_banlist:
            param_str += str(kwargs[key])

    return param_cache_path, data_cache_path, param_str


def create_cache(data, cache_folder, **kwargs):
    """Wrapper for making custom caches (no overwrite protection).

    * `data`: data to save
    * `cache_folder`: where to save
    * `**kwargs`: keyworded arguments to use to determine whether to load from
      cache later
        * If the same arguments aren't passed on load then an error will be
          raised (can be suppressed if desired)
    """

    # Make directory if it doesn't exist
    if cache_folder != "":
        try:
            os.makedirs(cache_folder)

        except OSError as e:  # pragma: no cover
            if e.errno != errno.EEXIST:
                raise

    param_cache_path, data_cache_path, cur_params = _make_cache_params(
        cache_folder, **kwargs
    )

    with open(param_cache_path, "wb") as handle:
        pickle.dump(cur_params, handle, protocol=pickle.HIGHEST_PROTOCOL)

    with open(data_cache_path, "wb") as handle:
        pickle.dump(data, handle, protocol=pickle.HIGHEST_PROTOCOL)


def load_cache(cache_folder, ignore_parameters=False, **kwargs):
    """Wrapper for loading custom caches.

    * `cache_folder`: where to load from
    * `ignore_parameters`: if `True` then suppress load error if `**kwargs` are
      different to when cache was saved
    * `**kwargs`: keyworded arguments to use to determine whether to load from
      the cache
    """
    param_cache_path, data_cache_path, cur_params = _make_cache_params(
        cache_folder, **kwargs
    )

    # Check that cache exists
    if not os.path.isfile(param_cache_path) or not os.path.isfile(data_cache_path):
        raise FileNotFoundError("Cannot find the cache ({})".format(cache_folder))

    with open(param_cache_path, "rb") as handle:
        cache_params = pickle.load(handle)

    if (cache_params == cur_params) or ignore_parameters:
        with open(data_cache_path, "rb") as handle:
            data_cache = pickle.load(handle)

        return data_cache

    else:
        raise RuntimeError(
            "Parameters changed, use 'ignore_parameters=True' to load anyway.\n\n"
            + "Parameters passed: {}\n\n".format(cur_params)
            + "Parameters in cache: {}".format(cache_params)
        )


def load_folder(
    data_path,
    file_extension=".csv",
    sep=",",
    skip_rows=None,
    header_row=0,
    target_cols=None,
    date_col=None,
    date_format=None,
    sort_by_date=False,
    min_duration=None,
    max_duration=None,
    cols_to_drop=None,
    drop_nan_cols=False,
    drop_static_cols=False,
    trim_data=False,
    trim_values=None,
    banlist=None,
    load_from_cache=False,
    save_to_cache=False,
    override_cache=False,
    cache_dir="",
):
    """Load all files from a folder into a Pandas DataFrame.

    Recommend use is indirectly via load_data().

    Loads all files in `data_path` that end with `file_extension` and treat
    each as a csv-alike file with separator `sep`. Assumes that all files follow
    the same format.

    * `data_path`: folder path to load from
    * `file_extension`: load files with this ending
    * `sep`: what to split the columns of data on e.g. `","` for `.csv`s or
      `"\t"` for `.tsb`s, not necessary for h5/parquet files
    * `skip_rows`: list of row indices in file to skip e.g., due to multi-row
      header, not necessary for h5/parquet files
    * `header_row`: which row contains header information, use `None` if no
      headers in the files, not necessary for h5/parquet files
    * `target_cols`: list of columns to load (others will be ignored), keep as `None` to
      load all columns
    * `date_col`: set to the index of a column with datetime information to
      force convert the column into a datetime
        - Use `date_format` in case of any peculiarity in the string
    * `date_format`: explicitly define expected datetime format for `date_col`
      if needed
        - Information on datetime string formatting can be found [here](https://docs.python.org/3/library/datetime.html#strftime-and-strptime-behavior)
    * `sort_by_date`: if `True` then each file will be stacked in ascending
      datetime order based on the first entry in `date_col`
        - `date_col` must be defined
    * `min_duration`: the minimum number of rows in a file (after other cleanup)
      for it to be added to the final dataframe
    * `max_duration`: the maximum number of rows in a file (after other cleanup)
      for it to be added to the final dataframe
    * `cols_to_drop`: list of column names to drop
    * `drop_nan_cols`: if `True`, remove any column that contains any
      `NaN` values
    * `drop_static_cols`: if `True`, drop columns that don't change throughout
      the entire loaded dataset
    * `trim_data`: if `True`, trim start and end of all data if any column
      matches any value in `trim_values`, defaults to `[np.NaN, -1, -55555, -99999]`
    * `trim_values`: which values to be considered for trimming
        - **Note** choosing bad values may simply remove all the data e.g. 0 if
          a signal is off in a particular file
    * `banlist`: list of filename strings to ignore during load
    * `load_from_cache`: if `True`, attempt to load from a cache located in
      `cache_dir`
    * `save_to_cache`: if `True`, save a cache into `cache_dir` if data hasn't
      been loaded from the cache
    * `override_cache`: if `True`,load even if `load_from_cache=True` and
      parameters differ
        - This will overwrite the previous cache if `save_to_cache=True`
    * `cache_dir`: what directory to save/load caches to/from

    Returns:

    * `df`: the loaded data
    * `file_path_lists`: list of file paths
    * `data_path`: from inputs
    """
    if skip_rows is None:
        skip_rows = []

    if cols_to_drop is None:
        cols_to_drop = []

    if banlist is None:
        banlist = []

    if trim_values is None:
        trim_values = [np.NaN, -1, -55555, -99999]

    saved_args = locals()
    # Load from cache
    if load_from_cache:
        try:
            data_cache = load_cache(
                cache_dir, ignore_parameters=override_cache, **saved_args
            )

            print("Found cache with matching parameters, loaded from cache.")

            if save_to_cache:
                print("Resaving, parameters may be updated..")
                create_cache(data_cache, cache_dir, **saved_args)

            return data_cache[0], data_cache[1], data_cache[2]

        except:
            print("Cache unavailable, loading from source..")

    if data_path == "":  # Catch local directory usage (breaks os.walk())
        data_path = os.getcwd()

    all_file_paths = [
        os.path.join(root, filename)
        for root, _, filenames in os.walk(data_path)
        for filename in filenames
        if filename.endswith(file_extension)
    ]

    if len(all_file_paths) == 0:
        raise FileNotFoundError("No files found in {}".format(data_path))

    # File order depends on filesystem therefore sort into lexicographical order to make
    # sure that it is consistent
    all_file_paths.sort()

    # Remove banlisted files
    file_names = []
    for path in all_file_paths:
        file_names.append(os.path.basename(path))

    all_file_paths = [
        x for i, x in enumerate(all_file_paths) if file_names[i] not in banlist
    ]

    # Read all the files into a list of dataframes
    if any(x in file_extension for x in ["h5", "h5py", "hdf5", "he5", "hdf"]):
        data_list = [pd.read_hdf(fp, columns=target_cols) for fp in all_file_paths]

    elif "parquet" in file_extension:  # req. pyarrow
        data_list = [
            pd.read_parquet(fp, engine="pyarrow", columns=target_cols)
            for fp in all_file_paths
        ]

    else:  # Otherwise assume a csv variant
        data_list = [
            pd.read_csv(
                fp,
                index_col=None,
                header=header_row,
                skiprows=skip_rows,
                sep=sep,
                usecols=target_cols,
            )
            for fp in all_file_paths
        ]

    # Rearrange files by their starting datetime
    if sort_by_date:
        if date_col is None:
            raise AttributeError("Must designate `date_col` to sort by date.")

        start_datetime_list = []
        for df in data_list:
            start_datetime_list.append(
                pd.to_datetime(df.iloc[0, date_col], format=date_format)
            )
        sorted_idxs = np.argsort(start_datetime_list)

        data_list = [data_list[i] for i in sorted_idxs]
        all_file_paths = [all_file_paths[i] for i in sorted_idxs]

    # Trim out artefacts
    if trim_data:
        for i in range(len(data_list)):
            # Find initial artefacts - loop rows until find a row with no artifacts
            for start_idx in range(data_list[i].shape[0]):
                if not data_list[i].iloc[start_idx, :].isin(trim_values).any():
                    break

            # Find end artefacts
            for end_idx in range(data_list[i].shape[0] - 1, 0, -1):
                if not data_list[i].iloc[end_idx, :].isin(trim_values).any():
                    break

            # Actual Trim
            data_list[i] = data_list[i].iloc[start_idx : end_idx + 1, :]

    # Concatenate into a single large dataframe
    df = pd.concat(data_list, keys=range(len(all_file_paths)))

    # Rename indices
    df.index.rename(["file_idx", "row_idx"], inplace=True)

    ### Top-level clean up
    # Convert to datetime object
    if date_col is not None:
        df.iloc[:, date_col] = pd.to_datetime(df.iloc[:, date_col], format=date_format)

    # Remove specified columns
    if len(cols_to_drop) >= 1:
        df.drop(cols_to_drop, axis=1, inplace=True)

    # Remove columns with NaNs in
    if drop_nan_cols:
        df.dropna(axis=1, inplace=True)

    # Remove columns that don't change
    if drop_static_cols:
        cols_to_remove = []
        for col in df:
            nb_uniques = df[col].unique().shape[0]
            if nb_uniques == 1 or math.isnan(nb_uniques):
                cols_to_remove.append(col)

        df.drop(cols_to_remove, axis=1, inplace=True)

    # Remove short flights, here rather than on df load to preserve file_idx association
    if min_duration is not None:
        final_mask = np.ones((df.shape[0],), dtype=np.bool)
        for flight_idx in df.index.unique(0):
            flight_mask = df.index.get_level_values(0).isin([flight_idx])

            flight_len = np.sum(flight_mask)

            if flight_len <= min_duration:
                final_mask = np.logical_xor(final_mask, flight_mask)

        df = df[final_mask]

    if max_duration is not None:
        final_mask = np.ones((df.shape[0],), dtype=np.bool)
        for flight_idx in df.index.unique(0):
            flight_mask = df.index.get_level_values(0).isin([flight_idx])

            flight_len = np.sum(flight_mask)

            if flight_len > max_duration:
                final_mask = np.logical_xor(final_mask, flight_mask)

        df = df[final_mask]

    ### Caching
    if save_to_cache:
        create_cache([df, all_file_paths, data_path], cache_dir, **saved_args)

    return df, all_file_paths, data_path


# Wrapper for loading multiple datasets (multiple folders)
def load_data(
    data_path_list,
    load_from_cache=False,
    save_to_cache=False,
    override_cache=False,
    cache_dir="",
    **kwargs
):
    """Wrapper for loading 1 or more datasets (folders) at the same time.

    Accepts all keyword arguments that `load_folder()` uses plus:

    * `data_path_list`: path or list of paths to folder(s) from which to load, replaces
      data_path from load_folder
    * `load_from_cache`: if `True` then try to load from `cache_dir`
    * `save_to_cache`: if `True` then try to save to `cache_dir`
    * `override_cache`: if `True` load even if `load_from_cache=True` and parameters
      differ
        - This will overwrite the previous cache if `save_to_cache=True`
    * `cache_dir`: what directory to save/load caches to/from

    Returns:

    * `df`: the loaded data
    * `file_path_lists`: list of file paths
    * `data_path_list`: from inputs

    Example usage:
    ```python
    import utc.loader

    # Load data from a single folder
    utc.loader.load_data(
        ["path/to/folder"],
    )

    # Load from multiple folders where other files are mixed in with target .csvs
    # Drop any files with <2000 entries and any columns that don't change throughout the
    # data as simple preprocessing
    utc.loader.load_data(
        ["path/to/folder1", "path/to/folder2"],
        file_extension=".csv",
        min_duration=2000,
        drop_static_cols=True,
    )
    ```

    """
    if not isinstance(data_path_list, list):
        data_path_list = [data_path_list]

    if load_from_cache:
        data_cache = load_cache(cache_dir, ignore_parameters=override_cache, **kwargs)

        print("Found cache with matching parameters, loaded from cache.")

        if save_to_cache:
            print("Resaving, parameters may be updated..")
            create_cache(data_cache, cache_dir, **kwargs)

        return data_cache[0], data_cache[1], data_cache[2]

    df_list = []
    file_path_lists = []
    for data_path in data_path_list:
        # Only pass kwargs not to do with caching
        df, file_paths, _ = load_folder(data_path, **kwargs)
        df_list.append(df)
        file_path_lists.append(file_paths)

    df = pd.concat(df_list, keys=range(len(file_path_lists)), sort=False)

    df.index.rename(["group_idx", df.index.names[1], df.index.names[2]], inplace=True)

    if save_to_cache:
        create_cache([df, file_path_lists, data_path_list], cache_dir, **kwargs)

    return df, file_path_lists, data_path_list


def get_meta_data(
    df, file_paths, extra_cols=None, extra_col_aliases=None, extra_col_locs=None
):
    """Returns information on each file in a loaded dataset

    * `df`: a DataFrame of the format produced by `load_data()`
    * `file_paths`: list that corresponds to the index in `df` as produced by
      `load_data`
    * `extra_cols`: list of columns to get a value from for each file (defaults to first
        value)
      e.g. for getting the starting timestamp for each file
    * `extra_col_aliases`: list of names to rename `extra_cols` columns to
    * `extra_col_locs`: list of integer locations to extra value from for `extra_cols`
        e.g. `0` for first value, `-1` for last value. Will default to `0`

    """

    # Check how many lists of lists we have in file_paths
    file_paths_list_depth = 0
    cur_val = file_paths
    while isinstance(cur_val, list) and len(cur_val) != 0:
        cur_val = cur_val[0]
        file_paths_list_depth += 1

    if file_paths_list_depth == 1:
        multi_folder = False
    elif file_paths_list_depth == 2:
        multi_folder = True
    else:
        raise RuntimeError(
            "Unexpected file_paths structure, should be list or list of lists "
        )

    if extra_cols is not None:
        if not isinstance(extra_cols, list):
            extra_cols = [extra_cols]

        if extra_col_aliases is not None:
            if not isinstance(extra_col_aliases, list):
                extra_col_aliases = [extra_col_aliases]
        else:
            extra_col_aliases = extra_cols

        if extra_col_locs is not None:
            if not isinstance(extra_col_locs, list):
                extra_col_locs = [extra_col_locs]
        else:
            extra_col_locs = [0 for x in range(len(extra_cols))]

        if len(extra_cols) != len(extra_col_aliases) or len(extra_cols) != len(
            extra_col_locs
        ):
            raise AttributeError(
                "extra_cols, extra_col_aliases and extra_col_locs must all be the same"
                + " length or be None/unspecified"
            )

    file_level_indices = sorted(set([x[:-1] for x in df.index]))

    lengths = []
    folders = []
    filenames = []
    for file_index in file_level_indices:
        lengths.append(df.loc[(file_index)].shape[0])

        if multi_folder:
            folder_path, filename = os.path.split(
                file_paths[file_index[0]][file_index[1]]
            )
        else:
            folder_path, filename = os.path.split(file_paths[file_index[0]])

        folders.append(os.path.basename(folder_path))
        filenames.append(filename)

    index = pd.MultiIndex.from_tuples(file_level_indices, names=df.index.names[:-1])
    meta_df = pd.DataFrame(
        {"folder": folders, "filename": filenames, "length": lengths}, index=index
    )

    if extra_cols is not None:
        for col, col_alias, col_loc in zip(
            extra_cols, extra_col_aliases, extra_col_locs
        ):
            tmp_list = []
            for idx in meta_df.index:
                tmp_list.append(df.loc[idx][col].iloc[col_loc])

            meta_df[col_alias] = pd.Series(tmp_list, index=meta_df.index)

    return meta_df


def convert_rdf_zips_into_csvs(data_dir_path, tool_path):  # pragma: no cover
    """Utility function for converting `.rdf` files to `.csv`

    * Install `Data Replay 3.0.6 - 14th Mar 2015.msi` (from this repo)
    * `data_dir_path`:location of the folder with the `.zip` files
    * `tool_path`: where Data Replay was installed
        - e.g., `C:/EMS/DataReplay/RDF_converter`
    * Files will then be converted and placed in the original folder alongside
      the `.zip` files
    """
    for file in os.listdir(data_dir_path):
        if file.endswith(".zip"):
            input_ = '"' + os.path.join(directory, file) + '"'
            output_cmd_ = "/output="

            base = os.path.splitext(file)[0]
            file_out = '"' + os.path.join(directory, base + ".csv") + '"'

            os.system(tool_path + " " + input_ + " " + output_cmd_ + " " + file_out)


def describe_dataset(
    df, max_uniques=5, n_top_correlated=5, sort_cols=True, save_path=""
):  # pragma: no cover
    """Give an overview of loaded data - based around `load_data()` format, but should
    work on most DataFrames.

    Slow and memory hungry on big datasets due to hunting for correlations.

    Save into a `.html` file by default for posterity.

    * `df`: relevant DataFrame
    * `max_uniques`: maximum unique values to display
    * `n_top_correlated`: number of correlated values to display
    * `sort_cols`: if `True` then sort column names alphabetically
    * `save_path`: directory to save `.html` to, set to `None` to not save
    """
    from tabulate import tabulate

    ### Overview
    info_str = "<p>"
    info_str += "Data Overview"

    for i in range(df.index.nlevels):
        if i != df.index.nlevels - 1:
            unique_indices = set([x[: -(df.index.nlevels - i - 1)] for x in df.index])
            nb_uniques = len(unique_indices)
        else:
            nb_uniques = df.index.size
        info_str += "<p>Level {:<2d}: {:<10s} : {:,} entries</p>".format(
            i, df.index.names[i], nb_uniques
        )

    info_str += "</p>"

    ### Parameter / Signal Table
    col_names = df.columns.values
    if sort_cols:
        col_names = sorted(col_names)

    valid_corr_cols = get_col_names_with_dtype(
        df, [float, int, bool], ignore_cols_with_nan=True
    )

    pearson_corr_df = {}
    spearman_corr_df = {}
    for col1 in valid_corr_cols:
        pearson_corr_df[col1] = []
        spearman_corr_df[col1] = []
        for col2 in valid_corr_cols:
            if col1 == col2:
                spearman_corr_df[col1].append(np.NaN)
                pearson_corr_df[col1].append(np.NaN)
                continue

            # Reversed input order since going into column-wise dictionary but later
            # will be retrieved by row index
            coef, p = spearmanr(df[col2], df[col1])
            spearman_corr_df[col1].append(coef)

            coef, p = pearsonr(df[col2], df[col1])
            pearson_corr_df[col1].append(coef)

    corr_matrix = {}
    corr_matrix["pearson"] = pd.DataFrame(pearson_corr_df, index=valid_corr_cols)
    corr_matrix["spearman"] = pd.DataFrame(spearman_corr_df, index=valid_corr_cols)

    uniques = []
    nb_nans = []
    types = []
    ranges = []
    q_ranges = []
    pearson_correlated = []
    spearman_correlated = []
    corr_lists = {"pearson": [], "spearman": []}
    for col in col_names:
        types.append(str(df[col].dtype))
        nb_nans.append(df[col].isnull().sum())

        # Add range info
        if df[col].dtypes in [np.int, np.float]:
            min_val = df[col].min()
            max_val = df[col].max()
            range_val = max_val - min_val
            ranges.append(
                "[{:.3f}, {:.3f}]<br/>({:.3f})".format(min_val, max_val, range_val)
            )

            min_q_val = df[col].quantile(0.05)
            max_q_val = df[col].quantile(0.95)
            range_q_val = max_q_val - min_q_val
            q_ranges.append(
                "[{:.3f}, {:.3f}]<br/>({:.3f})".format(
                    min_q_val, max_q_val, range_q_val
                )
            )
        else:
            ranges.append("N/A")
            q_ranges.append("N/A")

        # Add unique values
        cur_uniques = df[col].unique()
        if cur_uniques.shape[0] <= max_uniques:
            tmp_str = "Unique values:<br/>"
            for val in cur_uniques:
                if cur_uniques.dtype == np.float64:
                    tmp_str += "{:.3f}<br/>".format(val)
                else:
                    tmp_str += "{}<br/>".format(val)

            uniques.append(tmp_str)
        else:
            count_data = df[col].value_counts()
            col_mode = count_data.index[0]
            col_mode_count = count_data.iloc[0]

            uniques.append(
                "{} values<br/>mode: {}<br/>count={}".format(
                    cur_uniques.shape[0], col_mode, col_mode_count
                )
            )

        # Calculate most pearson correlated
        for corr_type in ["pearson", "spearman"]:
            cur_corr_matrix = corr_matrix[corr_type]
            if col in cur_corr_matrix.columns:
                most_corr_series = cur_corr_matrix.loc[col].iloc[
                    cur_corr_matrix.loc[col].to_numpy().argsort()[::-1]
                ]
                most_corr_series = most_corr_series[~np.isnan(most_corr_series)]

                if most_corr_series.shape[0] >= n_top_correlated:
                    correlation_string = ""
                    for i in range(n_top_correlated):
                        compare_col = most_corr_series.index[i]
                        coeff = most_corr_series.iloc[i]
                        correlation_string += "{}: {:.3f}<br/>".format(
                            compare_col, coeff
                        )

                    corr_lists[corr_type].append(correlation_string)
                else:
                    corr_lists[corr_type].append("Results Invalid")

            else:
                corr_lists[corr_type].append("Column not suitable")

    table_data = {
        "Signal": col_names,
        "Data Type": types,
        "# NaN": nb_nans,
        "Range": ranges,
        "Quantile Range<br/>(5% - 95%)": q_ranges,
        "Unique Values": uniques,
        "Pearson Correlation (Linear)": corr_lists["pearson"],
        "Spearman Correlation (Monotonic)": corr_lists["spearman"],
    }
    table_str = tabulate(table_data, headers="keys", tablefmt="html")

    final_html = html_table_style + info_str + table_str

    if running_in_jupyter():
        from IPython.display import HTML, display

        display(HTML(final_html))

    if save_path is not None:
        with open(save_path, "w+") as f:
            f.write(final_html)


def describe_data_splits(
    df_dict, quantile_range=None, save_path=None
):  # pragma: no cover
    """Give an overview of data splits.

    * `df_dict`: dictionary of data splits, expects 1 DataFrame for each key
    * `quantile_range`: What range to display
    * `save_path`: if not `None` then save as `.html` in this folder
    """
    from tabulate import tabulate

    if quantile_range is None:
        quantile_range = [0.05, 0.95]

    info_str_list = []
    table_str_list = []
    for key in df_dict.keys():
        info_str = "<p>" + key + "</p>"

        unique_multi_indexs = df_dict[key].reset_index(level=-1).index.unique()
        nb_uniques = len(unique_multi_indexs)
        sample_uniques = list(unique_multi_indexs[:3])
        info_str += "<p>Unique files: {}, ({}..)</p>".format(nb_uniques, sample_uniques)
        info_str_list.append(info_str)

        nb_cols = len(list(df_dict[key].columns))
        min_vals = np.zeros((nb_cols,))
        max_vals = np.zeros((nb_cols,))
        mean_vals = np.zeros((nb_cols,))
        l_quantile = np.zeros((nb_cols,))
        u_quantile = np.zeros((nb_cols,))

        for i, col in enumerate(df_dict[key].columns):
            try:
                min_vals[i] = df_dict[key][col].min()
                max_vals[i] = df_dict[key][col].max()
                mean_vals[i] = df_dict[key][col].mean()
                l_quantile[i] = df_dict[key][col].quantile(quantile_range[0])
                u_quantile[i] = df_dict[key][col].quantile(quantile_range[1])

            except (TypeError, ValueError):
                min_vals[i] = np.NaN
                max_vals[i] = np.NaN
                mean_vals[i] = np.NaN
                l_quantile[i] = np.NaN
                u_quantile[i] = np.NaN

        table_str_list.append(
            tabulate(
                {
                    "Column": df_dict[key].columns,
                    "Min": min_vals,
                    "Max": max_vals,
                    "Mean": mean_vals,
                    "Quantile ({})".format(quantile_range[0]): l_quantile,
                    "Quantile ({})".format(quantile_range[1]): u_quantile,
                    "Quantile Range": u_quantile - l_quantile,
                },
                headers="keys",
                tablefmt="html",
                numalign="decimal",
            )
        )

    for i in range(len(info_str_list)):
        if running_in_jupyter():
            from IPython.display import HTML, display

            display(HTML(html_table_style + info_str_list[i] + table_str_list[i]))

    if save_path is not None:
        with open(save_path, "w+") as f:
            f.write(html_table_style)
            for i in range(len(info_str_list)):
                f.write(info_str_list[i] + table_str_list[i])
