from tensorflow.keras import backend as K
import numpy as np


def gaussian_nll(y_true, y_pred):
    """Multivariate Gaussian negative loglikelihood loss function for Keras.

    Implies diagonal covariance matrix. Assumes first half of predictions are means
    and second half are log(sigma).

    Arguments:
        y_true {tensor} -- [n_samples, n_dims] ground truth values
        y_pred {tensor} -- [n_samples, n_dims*2] predicted mu and then log(sigma) values

    Returns:
        float -- negative log-likelihood averaged over samples
    """
    n_dims = int(y_pred.shape[1] / 2)
    mu = y_pred[:, :n_dims]
    logsigma = y_pred[:, n_dims:]

    mse = -0.5 * K.sum(
        K.square((y_true - mu) / (K.exp(logsigma) + K.epsilon())), axis=1
    )
    sigma_trace = -K.sum(logsigma, axis=1)
    log2pi = -0.5 * n_dims * np.log(2 * np.pi)

    log_likelihood = mse + sigma_trace + log2pi

    return K.mean(-log_likelihood)


def mse_for_nll(y_true, y_pred):
    """Mean Square Error for use with gaussian_nll (only use mu values)

    Assumes only first half of y_pred is actual prediction

    Arguments:
        y_true {tensor} -- [n_samples, n_dims] ground truth values
        y_pred {tensor} -- [n_samples, n_dims*2] predicted mu and then log(sigma) values

    Returns:
        float -- mean squared error (across samples)
    """
    n_dims = int(y_pred.shape[1] / 2)
    return K.mean(K.sum(K.square(y_pred[:, :n_dims] - y_true), axis=1))
