import os
import re
import copy
import warnings
import numpy as np
import pandas as pd

from sklearn import linear_model
from scipy.stats import spearmanr
from sklearn.preprocessing import PolynomialFeatures

from ..core import running_in_jupyter, uprint

if running_in_jupyter():  # pragma: no cover
    from tqdm.notebook import tqdm
else:
    from tqdm import tqdm


class MinMaxQuantileScaler:
    """Extension of [SKLearn MinMaxScaler](https://scikit-learn.org/stable/modules/generated/sklearn.preprocessing.MinMaxScaler.html)

    Performs min-max normalise to a given `feature_range` based on the min-max
    values of the quantile range.

    Designed to be pickled and unpickled without dependence.

    NaNs are treated as missing values: disregarded in fit, and maintained in
    transform.

    * `feature_range`: target feature range
    * `quantile_range`: quantile to use for min/max operations (useful for avoiding
       outliers)
    * `ignore_invalid`: if True then ignore data with invalid ranges (these will not
       be scaled). Check `scaled_cols` to ensure desired columns are scaled.

    If used with pandas series/dataframes then the column names acted on will be
    stored for later alignment.
    """

    def __init__(
        self, feature_range=(0, 1), quantile_range=(0, 1), ignore_invalid=False
    ):
        self.feature_range = feature_range
        self.quantile_range = quantile_range
        self.scaled_cols = None
        self.ignore_invalid = ignore_invalid

    def fit(self, X, cols=None):
        """Compute the parameters to be used for later scaling.
        Parameters

        * `X`: array-like of shape (n_samples, n_features)
            The data used to compute the per-feature minimum and maximum
            used for later scaling along the features axis.
        * `cols`: optional list of column names, if specified then only these
            columns will be acted upon (including in future transform and
            inverse_transform calls)
        """
        if self.feature_range[0] >= self.feature_range[1]:
            raise ValueError(
                "Minimum of desired feature range must be smaller"
                " than maximum. Got {}.".format(self.feature_range)
            )

        if self.quantile_range[0] >= self.quantile_range[1]:
            raise ValueError(
                "Minimum of desired quantile range must be smaller"
                " than maximum. Got {}.".format(self.quantile_range)
            )

        if cols is not None:
            if not isinstance(cols, list):
                cols = [cols]

            if not isinstance(X, (pd.DataFrame)):
                raise ValueError("Cols may only be used with DataFrames.")

            X = X[cols]

        data_min = np.nanquantile(X, self.quantile_range[0], axis=0)
        data_max = np.nanquantile(X, self.quantile_range[1], axis=0)

        self.n_samples_seen_ = X.shape[0]

        data_range = data_max - data_min
        any_invalid = np.any(data_range == 0) or np.any(data_range == np.nan)
        if any_invalid and not self.ignore_invalid:
            if isinstance(X, (pd.Series, pd.DataFrame)):
                err_string = "Some columns are static or NaN:\n"
                for i, col in enumerate(X.columns):
                    if data_range[i] == 0 or data_range[i] == np.nan:
                        err_string += "{} range: {}\n".format(col, data_range[i])

                raise ValueError(err_string)

            else:
                raise ValueError(
                    "One or more data ranges are 0 or NaN: {}".format(data_range)
                )
        elif any_invalid and self.ignore_invalid:
            invalid_mask = (data_range == 0) | (data_range == np.nan)
            invalid_cols = X.columns[invalid_mask]
            X = X.drop(invalid_cols, axis=1)

            data_range = data_range[~invalid_mask]
            data_min = data_min[~invalid_mask]
            data_max = data_max[~invalid_mask]

        self.scale_ = (self.feature_range[1] - self.feature_range[0]) / data_range
        self.min_ = self.feature_range[0] - data_min * self.scale_
        self.data_min_ = data_min
        self.data_max_ = data_max
        self.data_range_ = data_range

        if isinstance(X, (pd.Series, pd.DataFrame)):
            self.scaled_cols = list(X.columns)

    def transform(self, X, cols=None, no_bias=False, inplace=False):
        """Scale features of X according to feature_range.

        * `X`: array-like of shape (n_samples, n_features) to scale
        * `cols`: optional list of column names, if X is a DataFrame and the scaler was
            fit to a DataFrame then only scale `cols`
        * `no_bias`: if True then don't add offset term (just scale data)
        * `inplace`: if True then operate on the input array otherwise make a deepcopy
        """
        if not inplace:
            X = copy.deepcopy(X)

        # Assume 1D arrays are a single "column" of data (one signal, many observations)
        if len(X.shape) == 1:
            X = X[:, np.newaxis]

        if cols is not None:
            if not isinstance(cols, list):
                cols = [cols]

            scaling_idxs = [self.scaled_cols.index(x) for x in cols]

            if isinstance(X, (pd.DataFrame)):
                X[cols] *= self.scale_[scaling_idxs]
                if not no_bias:
                    X[cols] += self.min_[scaling_idxs]

            else:
                X *= self.scale_[scaling_idxs]
                if not no_bias:
                    X += self.min_[scaling_idxs]

        elif self.scaled_cols is not None and isinstance(X, (pd.Series, pd.DataFrame)):
            X[self.scaled_cols] *= self.scale_
            if not no_bias:
                X[self.scaled_cols] += self.min_

        else:
            X *= self.scale_
            if not no_bias:
                X += self.min_

        return np.squeeze(X)

    def inverse_transform(self, X, cols=None, no_bias=False, inplace=False):
        """Undo the scaling of X according to feature_range.

        * `X`: array-like of shape (n_samples, n_features) to un-scale
        * `cols`: optional list of column names, if X is a DataFrame and the scaler was
            fit to a DataFrame then only scale `cols`
        * `no_bias`: if True then don't add offset term (just scale data)
        * `inplace`: if True then operate on the input array otherwise make a deepcopy
        """
        if not inplace:
            X = copy.deepcopy(X)

        # Assume 1D arrays are a single "column" of data (one signal, many observations)
        if len(X.shape) == 1:
            X = X[:, np.newaxis]

        if cols is not None:
            if not isinstance(cols, list):
                cols = [cols]

            scaling_idxs = [self.scaled_cols.index(x) for x in cols]

            if isinstance(X, (pd.DataFrame)):
                if not no_bias:
                    X[cols] -= self.min_[scaling_idxs]
                X[cols] /= self.scale_[scaling_idxs]

            else:
                if not no_bias:
                    X -= self.min_[scaling_idxs]
                X /= self.scale_[scaling_idxs]

        elif self.scaled_cols is not None and isinstance(X, (pd.Series, pd.DataFrame)):
            if not no_bias:
                X[self.scaled_cols] -= self.min_
            X[self.scaled_cols] /= self.scale_

        else:
            if not no_bias:
                X -= self.min_
            X /= self.scale_

        return np.squeeze(X)


def zoh_model(y_true, file_divs=None):
    """Zero-Order Hold prediction.

    * `file_divs`: if specified then don't hold over file boundaries
    """
    if file_divs is None:
        return y_true[1:], y_true[:-1]

    else:
        y_pred = []
        y_new_true = []
        for file_limits in file_divs:
            y_pred.append(y_true[file_limits[0] : file_limits[1] - 1])
            y_new_true.append(y_true[file_limits[0] + 1 : file_limits[1]])

        y_pred = np.squeeze(np.concatenate(y_pred))
        y_new_true = np.squeeze(np.concatenate(y_new_true))

        return y_new_true, y_pred


def sklearn_predict_generator(sk_model, seq_generator, poly_degree=None):
    """Predict an SKLearn model on a data generator.

    * `poly_degree`: if not `None`, then convert data to polynomial format with
      specified degree
    """
    if poly_degree is not None:
        poly = PolynomialFeatures(poly_degree)

    y_pred = []
    for i in tqdm(range(seq_generator.__len__())):
        x_batch, _ = seq_generator.__getitem__(i)

        if poly_degree is not None:
            x_batch_org_shape = x_batch.shape
            x_batch = poly.fit_transform(x_batch.reshape((x_batch_org_shape[0], -1)))

        y_pred.append(sk_model.predict(x_batch))

    return np.concatenate(y_pred)


def linear_regression_fit_generator(seq_generator, poly_degree=None):
    """Optimally fit a linear regression on a Keras sequence generator.

    Useful for out-of-core data you want to fit without resorting to SGD.
    Assumes that the generator provides each sample in the data exactly once.

    Returns an SKLearn LinearRegression model.
    """
    if poly_degree is not None:
        poly = PolynomialFeatures(poly_degree)

    x_first_batch, y_first_batch = seq_generator.__getitem__(0)

    if poly_degree is not None:
        x_first_batch = poly.fit_transform(x_first_batch)

    # Check expected shape of final coefficients
    if len(y_first_batch.shape) == 1:
        coef_shape = (1, -1)  # Predict a single output

    elif len(y_first_batch.shape) == 2:
        coef_shape = (y_first_batch.shape[1], -1)  # Predict multiple outputs

    else:  # pragma: no cover
        raise AttributeError("Unusual batch shape: {}".format(y_first_batch.shape))

    # Do first batch to ensure matrices are appropriately shaped
    gram_x_cul = np.dot(x_first_batch.T, x_first_batch)
    phi_y_cul = np.dot(x_first_batch.T, y_first_batch)

    # Training
    for i in tqdm(range(1, seq_generator.__len__())):
        x_batch, y_batch = seq_generator.__getitem__(i)

        if poly_degree is not None:
            x_batch_org_shape = x_batch.shape

            x_batch = poly.fit_transform(x_batch.reshape((x_batch_org_shape[0], -1)))

        gram_x_cul += np.dot(x_batch.T, x_batch)
        phi_y_cul += np.dot(x_batch.T, y_batch)

    # Final calculation
    theta, _, _, _ = np.linalg.lstsq(gram_x_cul, phi_y_cul, rcond=None)
    theta = theta.T

    # Assumes we have a bias term from poly features or similar
    model = linear_model.LinearRegression(fit_intercept=False)

    model.coef_ = np.reshape(theta, coef_shape)
    model.intercept_ = 0.0
    return model


def mlflow_get_model(run_obj, custom_objects=None, test_data=None):
    """Get a model from mlflow, optionally run `evaluate_model_from_flavour()`"""
    import mlflow.sklearn
    import mlflow.keras

    if custom_objects is None:
        custom_objects = {}

    model_flavours = mlflow_get_model_flavours(run_obj)

    model_path = os.path.join(run_obj.artifact_uri, "model")
    if "sklearn" in model_flavours:
        model = mlflow.sklearn.load_model(model_path)

    elif "keras" in model_flavours:
        model = mlflow.keras.load_model(model_path, custom_objects=custom_objects)

    else:
        raise AttributeError(
            "No detected model flavours were loadable, found: {}".format(model_flavours)
        )

    if test_data is None:
        return model

    else:
        poly_degree = None
        if "params.degree" in run_obj.keys():
            poly_degree = int(run_obj["params.degree"])

        y_pred = evaluate_model_from_flavour(
            model, test_data, model_flavours, poly_degree=poly_degree
        )
        return model, y_pred


def evaluate_model_from_flavour(
    model, test_data, model_flavour=None, poly_degree=None, verbose=1
):
    """Utility function to run a model given what flavour it is.

    Supports "sklearn" and "keras", works on raw data or a generator.

    * `model_flavour`: what type of model `model` is
    * `poly_degree`: for use with polynomial data on "sklearn" models tested on
      a generator
    """
    data_is_array = False  # Assume a generator of some sort unless we see an array
    if isinstance(test_data, (pd.Series, pd.DataFrame, np.ndarray)):
        data_is_array = True

    # Force list for ease of use
    if not isinstance(model_flavour, list):
        model_flavour = [model_flavour]

    if "keras" in model_flavour:  # pragma: no cover (literally just keras calls)
        y_pred = np.squeeze(model.predict(test_data, verbose=verbose))

    elif "sklearn" in model_flavour:
        if data_is_array and poly_degree is None:
            y_pred = np.squeeze(model.predict(test_data))
        elif not data_is_array:
            y_pred = np.squeeze(
                sklearn_predict_generator(model, test_data, poly_degree=poly_degree)
            )

        else:
            raise RuntimeError("SKLearn model could not be run")

    else:
        raise RuntimeError(
            "Model flavour not understood, received: {}".format(model_flavour)
        )

    return y_pred


def mlflow_get_model_flavours(run_obj):
    """Get the flavours of model available for a given run as a list."""
    import mlflow

    model_path = os.path.join(run_obj.artifact_uri, "model")

    artifact_location = re.sub("file://", "", run_obj.artifact_uri)  # Cut if exists

    if os.name == "nt":  # Fix for windows (remove leading slash)
        artifact_location = re.sub("^/", "", artifact_location)

    model_flavours = mlflow.models.Model.load(
        os.path.join(artifact_location, "model", "MLmodel")
    ).flavors.keys()

    return list(model_flavours)


def mlflow_get_run(expt_name, run_name):
    """Get a run by name from the MLFlow database."""
    import mlflow

    expt_obj = mlflow.get_experiment_by_name(expt_name)
    try:
        run_obj = mlflow.search_runs(
            expt_obj.experiment_id, "tags.mlflow.runName='{}'".format(run_name)
        ).iloc[0]
    except IndexError as e:  # No run found
        raise RuntimeError("Run Not Found.")

    return run_obj


def mlflow_run_exists(expt_name, run_name):  # pragma: no cover
    """Return `True` if a run exists in the MLFlow database, `False` otherwise."""
    run_obj = mlflow_get_run(expt_name, run_name)

    if run_obj is None:
        return False
    else:
        return True


def get_spearman_ranks(
    df, target_col, other_cols=None, save_path=None, floatfmt=".3f"
):  # pragma: no cover
    """Get Spearman Rank Coefficients for `other_cols` predicting `target_col`.

    - `df`: input dataframe
    - `target_col`: column to compare to each of `other_cols`
    - `other_cols`: list of columns to compare to `target_col`
    - `save_path`: if not `None` then save html table to path
    - `floatfmt`: passed to tabulate, see [docs](https://pypi.org/project/tabulate/)
    """
    from tabulate import tabulate

    if other_cols is None:
        other_cols = [x for x in df.columns if x != target_col]

    if not isinstance(other_cols, list):
        other_cols = [other_cols]

    coefs = []
    p_vals = []
    valid_cols = []
    for col in other_cols:
        try:
            coef, p = spearmanr(df[col], df[target_col])
            coefs.append(coef)
            p_vals.append(p)
            valid_cols.append(col)
        except:
            warnings.warn(
                "Column '{}' could not be compared".format(col), RuntimeWarning
            )

    # Sort into descending order
    sorted_idxs = np.argsort(np.abs(coefs))[::-1]

    valid_cols = np.array(valid_cols)[sorted_idxs]
    coefs = np.array(coefs)[sorted_idxs]
    p_vals = np.array(p_vals)[sorted_idxs]

    table_data = {
        "{} vs".format(target_col): valid_cols,
        "Spearman Coeff.": coefs,
        "p": p_vals,
    }
    corr_table = tabulate(
        table_data, headers="keys", tablefmt="html", floatfmt=floatfmt
    )

    if running_in_jupyter():
        from IPython.display import HTML, display

        display(HTML(corr_table))

    if save_path is not None:
        with open(save_path, "w+") as f:
            f.write(corr_table)

    return corr_table


def check_tf_device_status():  # pragma: no cover
    """Prints basic hardware debug info for Tensorflow"""
    import tensorflow as tf

    print("Tensorflow Ver: {}\n".format(tf.__version__))
    print("Available Devices (Tensorflow):")
    for device in tf.python.client.device_lib.list_local_devices():
        print("\t{}".format(device.name))
