from tensorflow.keras.layers import *
from tensorflow.keras.models import Model

import numpy as np
import math


def tts_basic(
    input_shape,
    nb_temporal_filters=64,
    nb_spatial_filters=64,
    nb_dense_layers=2,
    nb_hidden_units=64,
    dropout_rate=0.5,
    lrelu_alpha=0.001,
):  # pragma: no cover
    """Parameterised generic Temporal-to-Spatial network design
    """
    inputs = Input(shape=input_shape)
    x = inputs

    # Temporal Convolutions
    x = Conv2D(
        nb_temporal_filters,
        (input_shape[0], 1),
        strides=(input_shape[0], 1),
        padding="valid",
        name="temporal_fold",
    )(x)
    x = LeakyReLU(lrelu_alpha)(x)

    # Spatial Convolutions
    x = Conv2D(
        nb_spatial_filters,
        (1, input_shape[1]),
        strides=(1, 1),
        padding="valid",
        name="spatial_fold",
    )(x)
    x = LeakyReLU(lrelu_alpha)(x)

    # Dense Layers
    x = Flatten()(x)
    for i in range(nb_dense_layers):
        x = Dense(nb_hidden_units, name="dense_layer_{}".format(i))(x)
        x = LeakyReLU(lrelu_alpha)(x)
        x = Dropout(rate=dropout_rate)(x)

    x = Dense(1, name="output_layer")(x)

    return Model(inputs, x)


def tts_skip_connections(
    input_shape,
    nb_fire_filters=16,
    nb_temporal_filters=64,
    nb_spatial_filters=64,
    nb_dense_layers=2,
    nb_hidden_units=128,
    dropout_rate=0.5,
    lrelu_alpha=0.001,
):  # pragma: no cover
    """Parameterised Temporal-to-Spatial network design with skip connections
    """
    inputs = Input(shape=input_shape)
    x = inputs

    # Fire module
    t0 = Conv2D(nb_fire_filters, (1, 1), strides=(1, 1), padding="same")(x)
    t0 = LeakyReLU(lrelu_alpha)(t0)

    t1 = Conv2D(nb_fire_filters, (3, 1), strides=(1, 1), padding="same")(t0)
    t1 = LeakyReLU(lrelu_alpha)(t1)

    t2 = Conv2D(nb_fire_filters, (1, 1), strides=(1, 1), padding="same")(t0)
    t2 = LeakyReLU(lrelu_alpha)(t2)

    f1 = Concatenate(axis=-1)([t1, t2])

    # Temporal Convolutions
    x = Conv2D(
        nb_temporal_filters,
        (input_shape[0], 1),
        strides=(input_shape[0], 1),
        padding="valid",
        name="temporal_fold",
    )(f1)
    x_temp = LeakyReLU(lrelu_alpha)(x)

    # Spatial Inception Module
    t1 = Conv2D(nb_spatial_filters, (1, 1), strides=(1, 1), padding="same")(x_temp)
    t1 = LeakyReLU(lrelu_alpha)(t1)
    t1 = Conv2D(
        nb_spatial_filters, (1, input_shape[1]), strides=(1, 1), padding="valid"
    )(t1)

    i_spatial = LeakyReLU(lrelu_alpha)(t1)

    # Dense Layers
    x = Flatten()(i_spatial)
    skip_input = Flatten()(inputs)
    skip_temporal = Flatten()(x_temp)
    skip_fire = Flatten()(f1)
    x = Concatenate(axis=-1)(
        [x, skip_input, skip_temporal, skip_fire]
    )  # Skip connections

    for i in range(nb_dense_layers):
        x = Dense(nb_hidden_units, name="dense_layer_{}".format(i))(x)
        x = LeakyReLU(lrelu_alpha)(x)
        x = Dropout(rate=dropout_rate)(x)

    x = Concatenate(axis=-1)([x, skip_input, skip_temporal, skip_fire])
    x = Dense(1, name="output_layer")(x)

    return Model(inputs, x)


def tts_confidence(
    input_shape,
    noise_sigma=None,
    nb_temporal_filters=64,
    temporal_conv_length=None,
    temporal_conv_strides=None,
    nb_spatial_filters=64,
    nb_mu_dense_layers=3,
    nb_sigma_dense_layers=3,
    nb_hidden_units=64,
    dropout_rate=0.5,
    lrelu_alpha=0.3,
):  # pragma: no cover
    """Reference implementation of the Temporal-to-Spatial network with confidence

    `noise_sigma`: if a number then a GaussianNoise layer is added at the input with the
        given noise sigma on all channels
    `nb_temporal_filters`: number of filters in the temporal convolutional layer
    `temporal_conv_length`: length of the temporal convolution, width=1. Defaults to the
        length of the input window
    `temporal_conv_strides`: stride of the temporal convolution, width=1. Defaults to
        the length of the input window
    `nb_spatial_filters`: number of filters in each component of the spatial fire module
    `nb_mu_dense_layers`: number of dense layers in the signal prediction path
    `nb_sigma_dense_layers`: number of dense layers in the signal confidence path
    `nb_hidden_units`: number of neurons in each of the aforementioned dense layers
    `dropout_rate`: dropout rate after each dense layer
    `lrelu_alpha`: alpha value for all LReLU activations
    """

    # Defaults
    if temporal_conv_length is None:
        temporal_conv_length = input_shape[0]

    if temporal_conv_strides is None:
        temporal_conv_strides = input_shape[0]

    true_inputs = Input(shape=input_shape)

    if noise_sigma is not None:
        inputs = GaussianNoise(noise_sigma)(true_inputs)
    else:
        inputs = true_inputs

    x_inputs_flat = Flatten()(inputs)
    x = inputs

    # Temporal Convolutions
    x = Conv2D(
        nb_temporal_filters,
        (input_shape[0], 1),
        strides=(temporal_conv_strides, 1),
        padding="valid",
        name="temporal_fold",
    )(x)
    x = LeakyReLU(lrelu_alpha)(x)

    # Spatial Fire Convolutions
    x = Conv2D(
        nb_spatial_filters, (1, 1), strides=(1, 1), padding="valid", name="spatial_s1x1"
    )(x)
    x_s1x1 = LeakyReLU(lrelu_alpha)(x)

    x = Conv2D(
        nb_spatial_filters,
        (1, input_shape[1]),
        strides=(1, 1),
        padding="valid",
        name="spatial_fold_1xN",
    )(x_s1x1)
    x_1xn = LeakyReLU(lrelu_alpha)(x)

    x = Conv2D(
        nb_spatial_filters,
        (1, 1),
        strides=(1, 1),
        padding="valid",
        name="spatial_fold_1x1",
    )(x_s1x1)
    x_1x1 = LeakyReLU(lrelu_alpha)(x)

    x_spatial_flat = Concatenate(axis=-1, name="spatial_concat")(
        [Flatten()(x_1x1), Flatten()(x_1xn)]
    )

    # Dense Layers
    x = Concatenate(axis=-1, name="inputs_and_tts")([x_spatial_flat, x_inputs_flat])
    for i in range(nb_mu_dense_layers):
        x = Dense(nb_hidden_units, name="mean_dense_{}".format(i))(x)
        x = LeakyReLU(lrelu_alpha)(x)
        x = Dropout(rate=dropout_rate)(x)

    z_mean = Dense(1, name="z_mean")(x)

    x = Concatenate(axis=-1, name="inputs_tts_pred")(
        [x_spatial_flat, x_inputs_flat, z_mean]
    )
    for i in range(nb_sigma_dense_layers):
        x = Dense(nb_hidden_units, name="logsigma_dense_{}".format(i))(x)
        x = LeakyReLU(lrelu_alpha)(x)
        x = Dropout(rate=dropout_rate)(x)

    z_std = Dense(1, name="z_std")(x)

    x = Concatenate(axis=-1)([z_mean, z_std])

    return Model(true_inputs, x)
