import bokeh.io
from bokeh.plotting import figure, show, ColumnDataSource, output_file
from bokeh.models import (
    Span,
    Label,
    DatetimeTickFormatter,
    Range1d,
    LinearAxis,
    WheelZoomTool,
    Patch,
)
from bokeh.models.tools import HoverTool
from bokeh.palettes import Category10_10, Category20_20, Greys3, viridis

import seaborn as sns

sns.set()

import matplotlib.pyplot as plt
import matplotlib.cm as cm  # For colour maps
import matplotlib.patches as patches
import matplotlib.gridspec as gridspec
from matplotlib.colors import Normalize as plt_normaliser


import os
import re
import copy
import warnings
import numpy as np
import pandas as pd

from ..core import slugify, get_nth_key

from sklearn.metrics import r2_score, mean_absolute_error, mean_squared_error
from pandas.plotting import register_matplotlib_converters

register_matplotlib_converters()


def plot_interactive(
    data,
    x_col=None,
    y_cols=None,
    y_axis_type="auto",
    y_range=None,
    y_label=None,
    y_cols2=None,
    y_label2=None,
    y_sigma=None,  # For plotting confidence intervals
    hide_cols=None,
    highlight_only=None,
    chart_type="line",
    x_label=None,
    x_range=None,
    x_datetime_format=None,
    legend_location=None,
    colour_palette=None,
    line_kwargs=None,
    circle_kwargs=None,
    extra_tooltips=None,
    title=None,
    plot_width=1400,
    plot_height=650,
    save_dir=None,
    notebook_url=None,
    supress_show=False,
):  # pragma: no cover
    """Make pretty, interactive plots with intelligent defaults.

    Every argument is optional (apart from the data), function will do its best to be
    nice anyway. Returns a handle to the figure.

    * `data`: may be a `pandas` DataFrame or series, or a dictionary of same
        - If DataFrame or Series then plot all `y_cols` and `y_cols2`
        - If dictionary then label data based on dictionary keys
    * `x_col`: what column to plot on the x axis
        - If not specified, will try to use the DataFrame index
    * `y_cols`: what column(s) to plot on the y axis
        - If not specified, will try to use all columns that aren't `x_col`
    * `y_axis_type`: set y axis type as per Bokeh e.g. "linear" or "log"
    * `y_range`: custom y range to use - useful for linking axes or setting limits
    * `y_label`: set y axis title
    * `y_cols2`: column(s) to plot on right hand side y axis
        - If not specified, will not plot anything
    * `y_label2`: set y2 axis title
    * `y_sigma`: column(s) to used to highlight a symmetrical area around the
      data in `y_cols` (pairs with y_cols 1:1), use for confidence highlights
    * `hide_cols`: hide these columns by default on the plot (can be turned on
      via legend)
    * `highlight_only`: these columns will show up, others will be grayed out and hidden
      from legends and tooltips
    * `chart_type`: what sort of plot to draw
        - Defaults to "line", may also be "circle" or "line_circle"
    * `x_label`: set x axis title
    * `x_range`: custom x range to use - useful for linking axes or setting limits
    * `x_datetime_format`: if specified then treat `x_col` as a datetime axis
        with the specified format
        - See [Bokeh docs](https://docs.bokeh.org/en/latest/docs/reference/models/formatters.html#bokeh.models.formatters.DatetimeTickFormatter) for info on specific formats
        - Generally follows [Python 3 datetime conventions](https://docs.python.org/3/library/datetime.html#strftime-and-strptime-behavior) e.g. `"%Y-%m-%d %H:%M:%S"`
    * `legend_location`: where to place the legend as per [docs](https://docs.bokeh.org/en/latest/docs/user_guide/styling.html?highlight=legend%20location%20top_right#location)
    * `colour_palette`: colour palette for plotting columns, see [Bokeh docs](https://docs.bokeh.org/en/latest/docs/reference/palettes.html)
    * `line_kwargs`: Bokeh formatting for "line" type plots, defaults to
        `{"line_width": 2}`
    * `circle_kwargs`: Bokeh formatting for "circle" type plots, defaults to
        `{"fill_alpha": 0.6, "line_width": 0, "size": 6}`
    * `extra_tooltips`: list of extra tooltips to use following [Bokeh docs](https://docs.bokeh.org/en/latest/docs/user_guide/tools.html#hovertool)
        - E.g. [("Flight Number", "@{flight}")] to add a tooltip that draws from
          the flight data column
    * `title`: what to title the plot, converted and used for saving as well
        - If not specified then one will be provided based off columns plotted
    * `plot_width`: graph width in pixels
    * `plot_height`: graph height in pixels
    * `save_dir`: if provided then the figure will be saved to this folder using
      the title to generate a filename
    * `notebook_url`: for live update - set to URL of notebook to allow
      `push_notebook` to work. Function will return a notebook handle in
      addition to the figure handle (see [official docs](https://docs.bokeh.org/en/latest/docs/user_guide/jupyter.html#notebook-handles))
        - Use `"default"` if running locally (`http://localhost:8888`)
        - Use `"guess"` if unsure
    * `supress_show`: if True then don't run show - useful for connecting figures

    If you want to create linked plots with unique colours:
    ```

    import utc.plotting
    from bokeh.io import show
    from bokeh.layouts import gridplot
    from bokeh.palettes import Category10_10

    import pandas as pd

    data = pd.DataFrame({
        "a": [1, 1, 1],
        "b": [2, 2, 2],
        "c": [3, 3, 3],
        "d": [4, 4, 4],
        "e": [5, 5, 5],
        "f": [6, 6, 6],
    })

    p1 = utc.plotting.plot_interactive(
        data,
        y_cols=["a", "b", "c"],
        colour_palette=Category10_10,
        supress_show=True,
    )

    p2 = utc.plotting.plot_interactive(
        data,
        y_cols=["d", "e", "f"],
        colour_palette=Category10_10[3:],
        x_range=p1.x_range,
        supress_show=True,
    )

    p = gridplot([[p1], [p2]])
    show(p)
    ```
    """
    if notebook_url is not None and supress_show:
        warnings.warn("supress_show does nothing when pushing live")

    if save_dir is not None and supress_show:
        warnings.warn("Note that files won't be saved until show() is called")

    if line_kwargs is None:
        line_kwargs = {"line_width": 2}

    if circle_kwargs is None:
        circle_kwargs = {"fill_alpha": 0.6, "line_width": 0, "size": 6}

    data, first_df, x_col, y_cols, y_cols2, title = _check_data_and_basic_inputs(
        data, x_col, y_cols, y_cols2, title
    )

    # Figure out notebook url
    if notebook_url is not None:
        if notebook_url == "default":
            notebook_url = "http://localhost:8888"

        # Pull running servers and take the first one
        elif notebook_url == "guess":
            from notebook import notebookapp

            notebook_url = list(notebookapp.list_running_servers())[0]["url"]

    # If y_sigma exists then make sure it is a list and has y_cols to apply to
    if y_sigma is not None:
        if not isinstance(y_sigma, list):
            y_sigma = [y_sigma]

        if len(y_sigma) > len(y_cols):
            raise AttributeError("There are more y_sigmas than y_cols to plot on")

    # Always make hide_cols a list for later searching (during plot)
    if not isinstance(hide_cols, list):
        hide_cols = [hide_cols]

    # Ensure highlight_only is a list
    if highlight_only is not None and not isinstance(highlight_only, list):
        highlight_only = [highlight_only]

    # Select a colour palette if one is not provided
    if colour_palette is None:
        max_colour_entries = 256
        if isinstance(data, dict):
            if highlight_only is not None:
                if len(highlight_only) > max_colour_entries:
                    nb_colour_entries = max_colour_entries
                    highlight_only = highlight_only[:max_colour_entries]

                    warnings.warn(
                        "Found too many highlights, plotting {}".format(
                            max_colour_entries
                        ),
                        UserWarning,
                    )
                else:
                    nb_colour_entries = len(highlight_only)

            else:
                nb_dict_keys = len(list(data.keys()))
                if nb_dict_keys > max_colour_entries:
                    raise AttributeError(
                        "Found too many dictionary entries, select <= {}".format(
                            max_colour_entries
                        )
                    )
                else:
                    nb_colour_entries = nb_dict_keys

        elif isinstance(data, pd.DataFrame):
            len_y2 = 0
            if y_cols2 is not None:
                len_y2 = len(y_cols2)

            nb_colour_entries = len(y_cols) + len_y2

            if nb_colour_entries > max_colour_entries:
                raise AttributeError(
                    "Found too many y_entries entries, select <= {}".format(
                        max_colour_entries
                    )
                )

        if nb_colour_entries <= 10:
            colour_palette = Category10_10
        elif nb_colour_entries <= 20:
            colour_palette = Category20_20
        else:
            colour_palette = viridis(nb_colour_entries)

    # Auto-detect y range if not set
    # Otherwise breaks with 2 y axes - uses range of second axis after plotting
    if y_range is None and len(y_cols) == 1 and y_cols2 is None:
        if first_df[y_cols[0]].dtype == np.dtype("O"):
            y_range = first_df[y_cols[0]].unique()

    elif y_range is None:
        min_val, max_val = get_extents_from_data(data, y_cols)
        y_range = Range1d(start=min_val, end=max_val)

    # Auto-detect categorical x axis
    if (
        x_range is None
        and not x_col == get_bokeh_index_col_name(first_df)
        and first_df[x_col].dtype == np.dtype("O")
    ):
        x_range = first_df[x_col].unique()

    # Get relevant columns of data to ensure sensible size of html
    relevant_cols = copy.deepcopy(y_cols)
    if x_col in first_df.columns:  # Catch where x_col is the index
        relevant_cols += [x_col]

    if isinstance(y_cols2, list):
        relevant_cols += y_cols2

    if isinstance(extra_tooltips, list):  # Grab what we want from tooltip strings
        for cur_tuple in extra_tooltips:
            matches = re.findall(r"@{*([^{}]*)}*", cur_tuple[1], re.MULTILINE)
            if len(matches) > 0:
                relevant_cols += [matches[0]]

    # Set up plot tools
    wheel_zoom_tool = WheelZoomTool()

    tools = ["pan", wheel_zoom_tool, "box_zoom", "undo", "redo", "reset", "save"]

    # Create hover tool
    hover_tool = HoverTool()

    if highlight_only is not None:
        hover_tool.names = highlight_only
        if x_datetime_format is not None:
            hover_tool.tooltips = [
                ("Source", "$name"),
                (y_cols[0], "@{{{}}}".format(y_cols[0])),
            ]
        else:
            hover_tool.tooltips = [
                ("Source", "$name"),
                (x_col, "@{{{}}}".format(x_col)),
                (y_cols[0], "@{{{}}}".format(y_cols[0])),
            ]

    else:
        if x_datetime_format is not None:
            hover_tool.tooltips = [("Name", "$name"), ("Value", "@$name")]
        else:
            hover_tool.tooltips = [
                ("Index", "@{{{}}}".format(x_col)),
                ("Name", "$name"),
                ("Value", "@$name"),
            ]

    # Append date if x column is a date
    if x_datetime_format is not None:
        hover_tool.tooltips.append(
            ("Date", "@{{{}}}{{{}}}".format(x_col, x_datetime_format))
        )
        hover_tool.formatters = {"@{{{}}}".format(x_col): "datetime"}

    if extra_tooltips is not None:
        if isinstance(extra_tooltips, list):
            for entry in extra_tooltips:
                hover_tool.tooltips.append(entry)
        else:
            raise AttributeError("extra_tooltips must be a list of tuples")

    tools.append(hover_tool)

    # Make Graph
    p = figure(
        tools=tools,
        toolbar_location="above",
        plot_width=plot_width,
        plot_height=plot_height,
        y_axis_type=y_axis_type,
        title=title,
        y_range=y_range,
        x_range=x_range,
        lod_threshold=1000,  # Make interactivity smoother
        lod_factor=10,
        lod_interval=150,
        lod_timeout=500,  # ms
    )

    # Set active tools
    # p.toolbar.active_scroll = wheel_zoom_tool

    # Set y axis label (must be here to avoid overwrite on y2 axis later)
    if y_label is not None:
        p.yaxis.axis_label = y_label
    elif len(y_cols) == 1:
        p.yaxis.axis_label = "{}".format(y_cols[0])
    else:
        p.yaxis.axis_label = "Amplitude"

    # Make extra axis if desired
    if y_cols2 is not None:
        min_val, max_val = get_extents_from_data(data, y_cols2)

        if y_label2 is None:
            y_label2 = ", ".join(y_cols2)

        p.extra_y_ranges = {"y2": Range1d(start=min_val, end=max_val)}
        p.add_layout(LinearAxis(y_range_name="y2", axis_label=y_label2), "right")

    # On a DataFrame use y_cols for legend labels
    if isinstance(data, pd.DataFrame):
        source = ColumnDataSource(data[relevant_cols])

        # Plot second axis if desired
        # Ordered so that axis one sits on top of axis two
        if y_cols2 is not None:
            _plot_on_bokeh_axes(
                plot=p,
                source=source,
                x_col=x_col,
                y_cols=y_cols2,
                colour_palette=colour_palette[len(y_cols) :],
                chart_type=chart_type,
                line_kwargs=line_kwargs,
                circle_kwargs=circle_kwargs,
                y_range_name="y2",
                hide_cols=hide_cols,
            )

        # Plot first axis
        _plot_on_bokeh_axes(
            plot=p,
            source=source,
            x_col=x_col,
            y_cols=y_cols,
            colour_palette=colour_palette,
            chart_type=chart_type,
            line_kwargs=line_kwargs,
            circle_kwargs=circle_kwargs,
            hide_cols=hide_cols,
        )

    # On a dict label legend based on dictionary keys
    if isinstance(data, dict):
        if highlight_only is not None:
            gray_keys = [key for key in data.keys() if key not in highlight_only]
            for key in gray_keys:
                source = ColumnDataSource(data[key][relevant_cols])
                _plot_on_bokeh_axes(
                    plot=p,
                    source=source,
                    x_col=x_col,
                    y_cols=y_cols,
                    colour_palette=Greys3,
                    chart_type=chart_type,
                    line_kwargs=line_kwargs,
                    circle_kwargs=circle_kwargs,
                    super_key="Other",
                    hide_cols=hide_cols,
                )

            for i, key in enumerate(highlight_only):
                source = ColumnDataSource(data[key][relevant_cols])
                _plot_on_bokeh_axes(
                    plot=p,
                    source=source,
                    x_col=x_col,
                    y_cols=y_cols,
                    colour_palette=colour_palette[i:],
                    chart_type=chart_type,
                    line_kwargs=line_kwargs,
                    circle_kwargs=circle_kwargs,
                    super_key=key,
                    super_name=key,
                    hide_cols=hide_cols,
                )

        # Otherwise plot normally - one colour per dictionary entry
        else:
            for i, key in enumerate(data.keys()):
                source = ColumnDataSource(data[key][relevant_cols])
                _plot_on_bokeh_axes(
                    plot=p,
                    source=source,
                    x_col=x_col,
                    y_cols=y_cols,
                    colour_palette=colour_palette[i:],
                    chart_type=chart_type,
                    line_kwargs=line_kwargs,
                    circle_kwargs=circle_kwargs,
                    super_key=key,
                    hide_cols=hide_cols,
                )

    # Plot sigma / confidence intervals, using patch (first y axis only)
    if y_sigma is not None:
        min_val = p.y_range.start
        max_val = p.y_range.end
        for i, (y_key, sigma_key) in enumerate(zip(y_cols, y_sigma)):
            # Use data source for index naming (in case passed a None)
            patch_x1 = source.data[x_col]
            patch_y1 = data[y_key] + data[sigma_key]

            # Mask off locations with NaN values in X (breaks plotting at start/end)
            # May not work correctly if NaNs in middle of data
            # Only use x1 & y1 since otherwise reverse may cause issue
            mask = np.full((patch_x1.shape[0],), True, dtype=np.bool)
            mask[np.isnan(patch_x1)] = False
            mask[np.isnan(patch_y1)] = False

            patch_x1 = patch_x1[mask]
            patch_y1 = patch_y1[mask]

            patch_x2 = patch_x1[::-1]
            patch_y2 = (data[y_key][mask] - data[sigma_key][mask])[::-1]

            x = np.hstack((patch_x1, patch_x2))
            y = np.hstack((patch_y1, patch_y2))

            patch_source = ColumnDataSource(dict(x=x, y=y))

            glyph = Patch(
                x="x", y="y", fill_color=colour_palette[i], fill_alpha=0.2, line_width=0
            )
            p.add_glyph(patch_source, glyph)

            min_y = patch_y2.min()
            if min_y < min_val:
                min_val = min_y

            max_y = patch_y1.max()
            if max_y > max_val:
                max_val = max_y

        # Update range
        p.y_range.start = min_val
        p.y_range.end = max_val

    # Figure out datetime
    # Set x axis label
    if x_label is not None:
        p.xaxis.axis_label = x_label

    else:
        p.xaxis.axis_label = x_col

    if x_datetime_format is not None:
        p.xaxis.formatter = DatetimeTickFormatter()

    # Legend properties
    if legend_location is not None:
        p.legend.location = legend_location

    p.legend.click_policy = "hide"

    # Save if desired
    if save_dir is not None:
        output_file(os.path.join(save_dir, slugify(title + ".html")), title=title)

    else:
        # Hack to remove any current file output without removing other outputs
        # e.g. notebook output
        current_state = bokeh.io.state.curstate()
        current_state._file = None  # Relies on knowing Bokeh internals
        # See https://docs.bokeh.org/en/latest/_modules/bokeh/io/state.html#State

    if notebook_url is None:
        if not supress_show:
            show(p)

        return p

    else:
        return p, show(p, notebook_url=notebook_url, notebook_handle=True)


def _check_data_and_basic_inputs(
    data, x_col, y_cols, y_cols2, title
):  # pragma: no cover
    """Ensure basic data is in a valid and usable format

    Ensures data is Series / DataFrame, the x_col argument is valid, and that y_cols and
    y_cols2 are lists as well as generating sensible defaults if None values are passed

    Arguments:
        data {DataFrame} -- data
        x_col {str, list} -- x column to use
        y_cols {str, list} -- y columns to plot (left axis)
        y_cols2 {str, list} -- y columns to plot (right axis)
        title {str} -- Figure title

    Returns:
        updated arguments -- data, first_df, x_col, y_cols, y_cols2, title
        first_df -- data or first dictionary entry for later reference

    """
    # Check data type
    if (
        not isinstance(data, dict)
        and not isinstance(data, pd.DataFrame)
        and not isinstance(data, pd.Series)
    ):
        raise AttributeError(
            "Data type {} not understood, use Pandas DataFrame or ".format(type(data))
            + "Series, alternatively use a dictionary of same."
        )

    data = copy.deepcopy(data)

    # Set up consistent access for later
    if isinstance(data, dict):
        # Coerce Series' to DataFrames
        for key in data.keys():
            if isinstance(data[key], pd.Series):
                data[key] = data[key].to_frame()

    elif isinstance(data, pd.Series):
        data = data.to_frame()

    # If using index as x_col then ensure sensible
    if x_col is None:
        if isinstance(data, dict):
            for key in data.keys():
                data[key] = data[key].reset_index()

            # Get from first (assume all DataFrames are the same)
            x_col = get_bokeh_index_col_name(data[get_nth_key(data, 0)])
            first_df = data[get_nth_key(data, 0)]

        elif isinstance(data, pd.DataFrame):
            data = data.reset_index()
            x_col = get_bokeh_index_col_name(data)
            first_df = data

    # Get "first" dataframe to calculate columns from etc (if only one dataframe then use that)
    if isinstance(data, dict):
        first_df = data[get_nth_key(data, 0)]

    elif isinstance(data, pd.DataFrame):
        first_df = data

    else:
        raise AttributeError("Unexpected type for input 'data'")

    # Check there is some actual data in the frame
    if first_df.shape[0] == 0 or first_df.shape[1] == 0:
        raise AttributeError("No data found")

    # If y_cols not defined then use all non-x_col columns
    if y_cols is None:
        y_cols = [x for x in first_df.keys() if x != x_col]

    # Convert non-list single y_col to list
    elif not isinstance(y_cols, list):
        y_cols = [y_cols]

    # Ensure that y_cols2 is a list
    if y_cols2 is not None and not isinstance(y_cols2, list):
        y_cols2 = [y_cols2]

    # But disallow empty list on y_cols2
    elif isinstance(y_cols2, list) and len(y_cols2) == 0:
        y_cols2 = None

    # Make a generic title if not specified
    if title is None:
        title = slugify("{} vs {}".format(x_col, y_cols))

    return data, first_df, x_col, y_cols, y_cols2, title


def get_bokeh_index_col_name(data):  # pragma: no cover
    # Get Bokeh source index column name via name wrangle (from Bokeh docs)
    if data.index.names[0] is None:
        index_col_name = "index"

    else:
        index_col_name = "_".join(data.index.names)

    return index_col_name


def _plot_on_bokeh_axes(
    plot,
    source,
    x_col,
    y_cols,
    colour_palette,
    chart_type,
    line_kwargs=None,
    circle_kwargs=None,
    y_range_name="default",
    super_key=None,
    super_name=None,
    hide_cols=None,
):  # pragma: no cover
    if line_kwargs is None:
        line_kwargs = {}

    if circle_kwargs is None:
        circle_kwargs = {}

    # Check valid chart type
    if chart_type not in ["line", "circle", "line_circle"]:
        raise AttributeError("Chart type {} not found.".format(chart_type))

    for i, key in enumerate(y_cols):
        if super_key is not None:
            legend_label = super_key
        else:
            legend_label = key

        if super_name is not None:
            name = super_name
        else:
            name = key

        if key in hide_cols:
            visible = False
        else:
            visible = True

        if chart_type == "line" or chart_type == "line_circle":
            line_handle = plot.line(
                x_col,
                key,
                y_range_name=y_range_name,
                color=colour_palette[i],
                source=source,
                legend_label=legend_label,
                name=name,  # Needed to use @name and $name
                visible=visible,
                **line_kwargs,
            )
        if chart_type == "circle" or chart_type == "line_circle":
            # Avoid double legend entries
            if chart_type == "line_circle":
                circle_handle = plot.circle(
                    x_col,
                    key,
                    y_range_name=y_range_name,
                    color=colour_palette[i],
                    source=source,
                    name=name,  # Needed to use @name and $name
                    visible=visible,
                    **circle_kwargs,
                )
            else:
                circle_handle = plot.circle(
                    x_col,
                    key,
                    y_range_name=y_range_name,
                    color=colour_palette[i],
                    source=source,
                    legend_label=legend_label,
                    name=name,  # Needed to use @name and $name
                    visible=visible,
                    **circle_kwargs,
                )


def get_extents_from_data(data, cols, extent_buffer=0.05):  # pragma: no cover
    """Given a dataframe and columns of interest get sensible axis limits
    * `extent_buffer`: how much space should be given either side of data (percentage)
    """
    if isinstance(data, pd.DataFrame):
        min_val = np.nanmin(data[cols].to_numpy(), axis=None)
        max_val = np.nanmax(data[cols].to_numpy(), axis=None)

    elif isinstance(data, dict):
        min_vals = []
        max_vals = []
        for key in data.keys():
            min_vals.append(np.nanmin(data[key][cols].to_numpy(), axis=None))
            max_vals.append(np.nanmax(data[key][cols].to_numpy(), axis=None))

        min_val = np.nanmin(min_vals)
        max_val = np.nanmax(max_vals)

    # Add buffer to extents
    if min_val < 0:
        min_val *= 1 + extent_buffer
    else:
        min_val *= 1 - extent_buffer

    if max_val < 0:
        max_val *= 1 - extent_buffer
    else:
        max_val *= 1 + extent_buffer

    return min_val, max_val


def plot_simple(
    data,
    x_col=None,
    y_cols=None,
    y_cols2=None,
    y_label=None,
    y2_axis_label=None,
    x_label=None,
    title=None,
    plot_width=18,
    plot_height=8,
    show_graph=True,
    return_fig=False,
    save_dir=None,
):  # pragma: no cover
    """Make nice non-interactive plots with intelligent defaults.
    * `data`: may be a `pandas` DataFrame or series
    * `x_col`: what column to plot on the x axis
        - If not specified, will try to use the DataFrame index
    * `y_cols`: what column(s) to plot on the y axis
        - If not specified, will try to use all columns that aren't `x_col`
    * `y_cols2`: column(s) to plot on right hand side y axis
        - If not specified, will not plot anything
    * `y_label`: label for `y_cols` axis
    * `y2_axis_label`: label for `y_cols2` axis
    * `x_label`: label for the x axis
    * `title`: what to title the plot, converted and used for saving as well
        - If not specified then one will be provided based off columns plotted
    * `plot_width`: width in matplotlib style
    * `plot_height`: height in matplotlib style
    * `show_graph`: if `True`, run `show()`
    * `return_fig`: if `True`, return figure handle
    """
    data, first_df, x_col, y_cols, y_cols2, title = _check_data_and_basic_inputs(
        data, x_col, y_cols, y_cols2, title
    )

    if x_col == get_bokeh_index_col_name(data):
        data.reset_index(level=-1, col_fill=x_col, inplace=True)

    # Actual Plotting
    fig, ax = plt.subplots(figsize=(plot_width, plot_height))

    # First axis
    for y_col in y_cols:
        ax.plot(data[x_col], data[y_col], label=y_col)

    # Optional second axis
    if y_cols2 is not None:
        ax2 = ax.twinx()

        for y_col in y_cols2:
            ax2.plot(data[x_col], data[y_col], label=y_col)

        if y2_axis_label is not None:
            ax2.set_ylabel(y2_axis_label)
        elif len(y_cols) == 1:
            ax2.set_ylabel("{}".format(y_cols[0]))
        else:
            ax2.set_ylabel("Amplitude")

    # Final additions and adjustments
    fig.legend()

    if y_label is not None:
        ax.set_ylabel(y_label)
    elif len(y_cols) == 1:
        ax.set_ylabel("{}".format(y_cols[0]))
    else:
        ax.set_ylabel("Amplitude")

    if x_label is not None:
        ax.set_xlabel(x_label)
    else:
        ax.set_xlabel(x_col)

    ax.set_title(title)
    fig.tight_layout()

    # Save if desired
    if save_dir is not None:
        fig.savefig(
            os.path.join(save_dir, slugify(title + ".png")), bbox_inches="tight"
        )

    if show_graph:
        plt.show(fig)

    if return_fig:
        return fig


def plot_heatmap(
    data,
    gap_indices=None,
    gap_size=30,
    mask=None,
    x_label=None,
    y_label=None,
    make_indices_str=False,
    make_cols_indices=False,
    title=None,
    plot_width=22,
    plot_height=18,
    save_dir=None,
    cmap="YlGnBu",
    return_fig=False,
    **kwargs,
):  # pragma: no cover
    """Wrap Seaborn's `heatmap`, allowing for some nice extensions
    * `data`: may be a `pandas` DataFrame or series
    * `gap_indices`: create gaps at the specified indices
    * `gap_size`: size of gap to create
    * `mask`: boolean array (mask) of what data to remove/hide
    * `x_label`: x axis label
    * `y_label`: y axis label
    * `make_indices_str`: if True, make indices strings for better plotting
    * `make_cols_indices`: if True, use column names as indices
    * `title`: plot title
    * `plot_width`: width of plot in `matplotlib` dimensions
    * `plot_height`: height of plot in `matplotlib` dimensions
    * `save_dir`: if provided then the figure will be saved to this folder using
      the title to generate a filename
    * `cmap`: colour plaette to use, [see here](https://seaborn.pydata.org/tutorial/color_palettes.html)
    * `return_fig`: if True, return figure handle for later manipulation
    * Supports all other arguments to [Seaborn's `heatmap`](https://seaborn.pydata.org/generated/seaborn.heatmap.html)
    """
    if make_cols_indices and make_indices_str and data.shape[0] != data.shape[1]:
        raise AttributeError(
            "make_cols_indices=True and make_indices_str=True is only supported for square dataframes."
        )

    if title is None:
        title = "Heatmap"

    if x_label is None:
        x_label = ""

    if y_label is None:
        y_label = "Index"

    if mask is None:
        mask = np.full(data.shape, False, dtype=np.bool)

    org_cols = data.columns

    if make_indices_str:
        org_df_indices = data.index
        new_index = []
        for i in range(data.shape[0]):
            new_index.append(str(org_df_indices[i]))
    else:
        new_index = data.index

    # Add gaps if relevant
    if gap_indices is not None:
        gap_locs = [data.index.get_loc(x) for x in gap_indices]
        gap_diff = np.diff([0] + gap_locs + [data.shape[0]])

        if make_cols_indices:
            new_dims = (
                data.shape[0] + gap_size * (len(gap_locs)),
                data.shape[1] + gap_size * (len(gap_locs)),
            )

        else:
            new_dims = (data.shape[0] + gap_size * (len(gap_locs)), data.shape[1])

        new_mask = np.full(new_dims, False, dtype=np.bool)
        new_data = np.zeros(new_dims)

        # Make a mask of valid indices
        valid_idx_mask = np.zeros((new_dims[0],), dtype=np.bool)
        start_loc = 0
        for nb_valid in gap_diff:
            end_loc = start_loc + nb_valid
            valid_idx_mask[start_loc:end_loc] = True
            start_loc = end_loc + gap_size

        # Update index
        updated_index = []
        org_index = 0
        for i in range(new_dims[0]):
            if valid_idx_mask[i]:
                updated_index.append(str(new_index[org_index]))
                org_index += 1

            else:
                updated_index.append("-")

        # Copy across relevant bits of mask and data
        i_val = 0
        j_val = 0
        for i in range(new_dims[0]):
            if not valid_idx_mask[i]:
                continue
            # On valid rows draw stuff
            for j in range(new_dims[1]):
                if not make_cols_indices or valid_idx_mask[j]:
                    new_data[i, j] = data.iloc[i_val, j_val]
                    new_mask[i, j] = mask[i_val, j_val]
                    j_val += 1

            j_val = 0
            i_val += 1

        new_mask[~valid_idx_mask, :] = True

        if make_cols_indices:
            new_mask[:, ~valid_idx_mask] = True

            final_data = pd.DataFrame(
                data=new_data, index=updated_index, columns=updated_index
            )

        else:
            final_data = pd.DataFrame(
                data=new_data, index=updated_index, columns=org_cols
            )

        mask = new_mask

    else:
        if make_cols_indices:
            final_data = pd.DataFrame(
                data=data.to_numpy(), index=new_index, columns=new_index
            )
        else:
            final_data = pd.DataFrame(
                data=data.to_numpy(), index=new_index, columns=org_cols
            )

    # Plotting
    fig, ax = plt.subplots(figsize=(plot_width, plot_height))
    ax = sns.heatmap(final_data, mask=mask, cmap=cmap, **kwargs)

    ax.set_title(title)
    ax.set_xlabel(x_label)
    ax.set_ylabel(y_label)

    if save_dir is not None:
        fig.savefig(
            os.path.join(save_dir, slugify(title + ".png")), bbox_inches="tight"
        )

    if return_fig:
        return fig


def prediction_performance_battery(
    y_true,
    y_pred,
    set_titles=None,
    joint_kind=None,
    subplot_width=7,
    subplot_height=7.5,
    fix_joint_limits=None,
    title=None,
    calc_metrics=None,
    save_dir=None,
):  # pragma: no cover
    """Performance battery for prediction models.

    Allows lists of `y_true` and `y_pred` e.g. `[y_train_pred, y_test_pred]` to
    make it possible to compare multiple data sets in one place.

    `y_true`, `y_pred` and `set_titles` (if used) must be synced lists.

    * `y_true`: list of arrays of the true values associated with a set e.g.
      your train/validation/test set
    * `y_pred`: list of arrays of the predicted values associated with a set,
      must be synced with `y_true` i.e. `y_true[0]` and `y_pred[0]` must be the
      true and predicted values for the same set, respectively
    * `joint_kind`: list of [seaborn jointplots](https://seaborn.pydata.org/generated/seaborn.jointplot.html)
      to plot, defaults to `["scatter", "hex"]`
    * `fix_joint_limits`: list `[x_min, x_max, y_min, y_max]` of custom axis
      limits for joint plot
    * `set_titles`: title prefix for each plot associated with the synced entry
      in `y_true`
    * `title`: super title to use for the plot as a whole
    * `calc_metrics`: list that can contain zero or more of ["mae", "mse", "r2"]
      for each entry the associated metric will be calculated and include in the
      distribution plot title
    """
    if joint_kind is None:
        joint_kind = ["scatter", "hex"]

    if calc_metrics is None:
        calc_metrics = []

    if title is None:
        title = "Model Performance Battery"

    if not isinstance(y_true, list):
        y_true = [y_true]

    if not isinstance(y_pred, list):
        y_pred = [y_pred]

    if set_titles is None:
        set_titles = []
        for i in range(len(y_pred)):
            set_titles.append("Set {}".format(i))

    elif set_titles is not None and not isinstance(set_titles, list):
        set_titles = [set_titles]

    if (len(y_pred) != len(y_true)) or (len(y_pred) != len(set_titles)):
        raise AttributeError("y_true, y_pred, and set_titles must be the same length")

    nb_plot_cols = len(joint_kind) + 1
    nb_plot_rows = len(y_pred)
    fig = plt.figure(
        figsize=(subplot_width * nb_plot_cols, subplot_height * nb_plot_rows)
    )
    gs = gridspec.GridSpec(nb_plot_rows, nb_plot_cols)

    # Joint Plots of True vs Predicted
    if fix_joint_limits is None:
        shared_xlims = [np.inf, -np.inf]
        shared_ylims = [np.inf, -np.inf]
    else:
        shared_xlims = [fix_joint_limits[0], fix_joint_limits[1]]
        shared_ylims = [fix_joint_limits[2], fix_joint_limits[3]]

    jointplot_list = []
    for cur_pred, cur_true, cur_set_title in zip(y_pred, y_true, set_titles):
        for i, cur_kind in enumerate(joint_kind):
            g = sns.jointplot(x=cur_true, y=cur_pred, kind=cur_kind, linewidth=0)
            jointplot_list.append(g)

            g.set_axis_labels(xlabel="True Value", ylabel="Predicted Value")

            # Update global axis limits
            if fix_joint_limits is None:
                x_min, x_max = g.fig.axes[0].get_xlim()
                y_min, y_max = g.fig.axes[0].get_ylim()

                if x_min < shared_xlims[0]:
                    shared_xlims[0] = x_min

                if x_max > shared_xlims[1]:
                    shared_xlims[1] = x_max

                if y_min < shared_ylims[0]:
                    shared_ylims[0] = y_min

                if y_max > shared_ylims[1]:
                    shared_ylims[1] = y_max

    # Update jointplot limits and add y=x line, then copy to main figure (fig)
    for i in range(nb_plot_rows):
        for j in range(len(joint_kind)):
            jointplot_index = i * len(joint_kind) + j
            cur_g_axis = jointplot_list[jointplot_index]

            # Plot y=x
            x1 = max(shared_xlims[0], shared_ylims[0])
            x2 = min(shared_xlims[1], shared_ylims[1])

            cur_g_axis.ax_joint.plot(
                [x1, x2], [x1, x2], "g--", linewidth=2, label="y=x"
            )
            cur_g_axis.ax_joint.legend()

            # Set axis limits
            cur_g_axis.ax_marg_x.set_xlim(shared_xlims[0], shared_xlims[1])
            cur_g_axis.ax_marg_y.set_ylim(shared_ylims[0], shared_ylims[1])

            gridspec_index = i * (nb_plot_cols) + j
            SeabornFig2Grid(cur_g_axis, fig, gs[gridspec_index])

            sub_title = "{} ".format(set_titles[i])
            if joint_kind[j] == "hex":
                sub_title += "Prediction Error Density"

            elif joint_kind[j] == "scatter":
                sub_title += "Prediction Error Scatter"

            else:
                sub_title += "Joint Plot Style: {}".format(joint_kind[j])

            # Take advantage of knowing how a jointplot uses sub-figures
            fig.axes[-2].set_title(sub_title)

    # Do residual plots last so hack with jointplot titles works
    for i in range(nb_plot_rows):
        # Normalised Residual Error Distribution
        cur_ax = fig.add_subplot(gs[(i * nb_plot_cols) + (nb_plot_cols - 1)])
        sns.histplot(y_true[i] - y_pred[i], ax=cur_ax)

        title_str = "{} Error Distribution".format(set_titles[i])
        for j, metric in enumerate(calc_metrics):
            if j == 0:
                title_str += "\n"
            else:
                title_str += ", "

            if metric == "mae":
                title_str += "mae: {:.03f}".format(
                    mean_absolute_error(y_true[i], y_pred[i])
                )
            elif metric == "mse":
                title_str += "mse: {:.03f}".format(
                    mean_squared_error(y_true[i], y_pred[i])
                )
            elif metric == "r2":
                title_str += "$r^2$: {:.03f}".format(r2_score(y_true[i], y_pred[i]))
            else:
                warnings.warn("Metric '{}' not understood".format(metric))

        cur_ax.set_title(title_str)

        cur_ax.set_xlabel("Error")
        cur_ax.set_ylabel("Probability Density")

    fig.suptitle(title)

    # Fix suptitle in tight layout
    title_offset = (0.06 / (nb_plot_rows)) + max(0, nb_plot_rows - 2) * 0.005
    gs.tight_layout(fig, rect=[0, 0, 1, 1 - title_offset])
    if save_dir is not None:
        fig.savefig(
            os.path.join(save_dir, slugify(title + ".png")), bbox_inches="tight"
        )


class SeabornFig2Grid:  # pragma: no cover
    """Copy a seaborn figure into a matplotlib grid e.g. side-by-side jointplots

    Shim from [Stack Overflow](https://stackoverflow.com/questions/35042255/how-to-plot-multiple-seaborn-jointplot-in-subplot).

    Allows moving of the multiple figures in a jointplot or similar to another
    matplotlib figure, e.g. if you want to plot several joint plots next to each
    other. This is surprisingly difficult.

    Examples usage:

    ```python
    nb_plots = 2
    fig = plt.figure(figsize=(7 * nb_plots, 8))
    gs = gridspec.GridSpec(1, nb_plots)

    g1 = sns.jointplot(x=some_x, y=some_y)
    SeabornFig2Grid(g1, fig, gs[0])

    g2 = sns.jointplot(x=some_other_x, y=some_other_y)
    SeabornFig2Grid(g2, fig, gs[1])
    ```
    """

    def __init__(self, seaborngrid, fig, subplot_spec):
        self.fig = fig
        self.sg = seaborngrid
        self.subplot = subplot_spec
        if isinstance(self.sg, sns.axisgrid.FacetGrid) or isinstance(
            self.sg, sns.axisgrid.PairGrid
        ):
            self._movegrid()
        elif isinstance(self.sg, sns.axisgrid.JointGrid):
            self._movejointgrid()
        self._finalize()

    def _movegrid(self):
        """ Move PairGrid or Facetgrid """
        self._resize()
        n = self.sg.axes.shape[0]
        m = self.sg.axes.shape[1]
        self.subgrid = gridspec.GridSpecFromSubplotSpec(n, m, subplot_spec=self.subplot)
        for i in range(n):
            for j in range(m):
                self._moveaxes(self.sg.axes[i, j], self.subgrid[i, j])

    def _movejointgrid(self):
        """ Move Jointgrid """
        h = self.sg.ax_joint.get_position().height
        h2 = self.sg.ax_marg_x.get_position().height
        r = int(np.round(h / h2))
        self._resize()
        self.subgrid = gridspec.GridSpecFromSubplotSpec(
            r + 1, r + 1, subplot_spec=self.subplot
        )

        self._moveaxes(self.sg.ax_joint, self.subgrid[1:, :-1])
        self._moveaxes(self.sg.ax_marg_x, self.subgrid[0, :-1])
        self._moveaxes(self.sg.ax_marg_y, self.subgrid[1:, -1])

    def _moveaxes(self, ax, gs):
        # https://stackoverflow.com/a/46906599/4124317
        ax.remove()
        ax.figure = self.fig
        self.fig.axes.append(ax)
        self.fig.add_axes(ax)
        ax._subplotspec = gs
        ax.set_position(gs.get_position(self.fig))
        ax.set_subplotspec(gs)

    def _finalize(self):
        plt.close(self.sg.fig)
        self.fig.canvas.mpl_connect("resize_event", self._resize)
        self.fig.canvas.draw()

    def _resize(self, evt=None):
        self.sg.fig.set_size_inches(self.fig.get_size_inches())


def conv_filter_visualisation_2d(
    filters,
    plot_width=20,
    highlight_index=None,
    x_label="Location",
    x_ticklabels=None,
    global_norm=False,
    title=None,
    save_dir=None,
):  # pragma: no cover
    """Visualise filters weights in a 2D convolution.

    * `filters`: layer weights
    * `highlight_index`: index to highlight with a red box
    * `global_norm`: whether to globally normalise
    * `x_ticklabel`: custom x tick labels
    * `save_dir`: directory to save into
    """
    if title is None:
        title = "Filter Visualisation"

    nb_filters = filters.shape[-1]
    len_filters = filters.shape[0]

    # Standardise to [0 1]
    if global_norm:
        f_min = filters.min()
        f_max = filters.max()
        norm_filters = (filters - f_min) / (f_max - f_min)
    else:
        norm_filters = copy.deepcopy(filters)
        for i in range(nb_filters):
            f_min = norm_filters[:, :, :, i].min()
            f_max = norm_filters[:, :, :, i].max()
            norm_filters[:, :, :, i] = (norm_filters[:, :, :, i] - f_min) / (
                f_max - f_min
            )

    colour_map = cm.ScalarMappable(norm=plt_normaliser(vmin=0.0, vmax=1.0), cmap="gray")

    fig, axes = plt.subplots(
        nb_filters + 1,
        1,
        figsize=(plot_width, (nb_filters + 1) * (1.25 * plot_width / len_filters)),
        sharex=True,
    )

    axes[0].set_title(title)
    for i in range(nb_filters):
        axes[i].set_yticks([])

        im = axes[i].imshow(
            norm_filters[:, :, 0, i].T, cmap=colour_map.get_cmap(), aspect="equal"
        )

        # Highlight prediction point
        if highlight_index is not None:
            _draw_highlight_index_matplotlib(axes[i], im, len_filters, highlight_index)

    # Add additional plot with average filter
    avg_filters = np.sum(filters, axis=-1)
    avg_f_min = avg_filters.min()
    avg_f_max = avg_filters.max()
    norm_avg_filters = (avg_filters - avg_f_min) / (avg_f_max - avg_f_min)

    axes[-1].set_title("Average Filter")
    axes[-1].set_xlabel(x_label)
    axes[-1].set_yticks([])

    if x_ticklabels is not None:
        axes[-1].set_xticks(np.array(range(len_filters)))
        axes[-1].set_xticklabels(x_ticklabels, rotation=90)

    axes[-1].imshow(
        norm_avg_filters[:, :, 0].T, cmap=colour_map.get_cmap(), aspect="equal"
    )

    if highlight_index is not None:
        _draw_highlight_index_matplotlib(axes[-1], im, len_filters, highlight_index)

    l, b, w, h = axes[-1].get_position().bounds

    axes[-1].set_position([l, b - (h * 2.1), w, h])

    # Add Colour Bar last so it knows about coordinate changes
    cbar = fig.colorbar(colour_map, ax=axes.ravel().tolist(), shrink=1.0)

    plt.show()

    if save_dir is not None:
        fig.savefig(
            os.path.join(save_dir, "{}.png".format(slugify(title))), bbox_inches="tight"
        )

    return fig


def _draw_highlight_index_matplotlib(
    ax, im, len_filters, highlight_index, linewidth=1, edgecolor="r"
):  # pragma: no cover
    extents = im.get_extent()

    x_start = ((extents[1] - extents[0]) / len_filters) * highlight_index + extents[0]
    side_len = extents[2] - extents[3]

    rect = patches.Rectangle(
        (x_start, extents[3]),
        side_len,
        side_len,
        linewidth=linewidth,
        edgecolor=edgecolor,
        facecolor="none",
    )
    ax.add_patch(rect)
