"""Module to help with windowing datasets, intended to be used with `loading`.

Functions and classes in this module assume that DataFrames provided that have
a Multi-Index as produced by `loading.load_data()` where if there are `N`
indices the highest `N-1` uniquely define a file/test/flight e.g. (`group_idx`,
`file_idx`, `row_idx`). I.e. only the final index is continuous and therefore
windows must not span boundaries in any of the top `N-1` indices. **Note** if
adding extra indices they should, therefore, be placed at the `-2` level.
"""

from tensorflow.keras.utils import Sequence
import numpy as np
import pandas as pd

from ..core import get_mask_from_indices


class MultiIndexDFWindowGenerator(Sequence):
    """Window generator class compatible with Keras

    Ensures that windows aren't generated over file boundaries.

    * `data`: should be a DataFrame with a multi-index where the lowest level
      represents a row in a file and the second lowest is a single file e.g.,
      created by `loader.load_data()`
    * `win_inc`: is the increment between successive windows and therefore
      controls the amount of overlap
    * `win_len`: controls the length of each window
    * `in_cols`: the column names to use as input data
    * `out_cols`: the column names to use as output data
    * `weight_col`: if specified then include weighting for each batch element
      based on this column
      Use alongside `weight_callback` 
    * `weight_callback`: defines how to match weighting information with a window.
      This function is called on the weight data column for each window. It defaults 
      to taking the most recent weight within the window
        - `w_data: w_data[int(WIN_LEN/2)]` would take the weight from middle of the
          window
        -  `w_data: np.mean(w_data)` would give the mean weight over a window
    * `x_shape=None`: tuple defining how to reshape the output when windowing
      e.g. to add extra dimensions, assumed to be `(win_len, len(in_cols))`
      otherwise
    * `out_callback`: the function to call on the output data before return, the
      default is to take all the data from the most recent timestep in each
      output column
        - `y_data: y_data[int(WIN_LEN/2)]` would take from the middle of the window
        - `y_data: y_data` would return the whole window of y data
    * `batch_size`: how big each batch should be
    * `allow_partial_final_batch`: if True then allow a smaller final batch so
      that all data is presented rather than skipping when the size of the data
      is not divisible by `batch_size`
        - Use for testing and validation data generators
    * `shuffle`: whether to shuffle the data (reshuffled after each epoch)
    * `win_augment_callback`: callback to apply to all windows in x batches
        * Must take one argument which is a window of data and return the augmented
            version
        * E.g. `lambda x_win: # something`
    """

    def __init__(
        self,
        df,
        in_cols=None,
        out_cols=None,
        weight_col=None,
        weight_callback=lambda w_data: w_data[-1],
        win_inc=1,
        win_len=1,
        x_shape=None,
        out_callback=lambda y_data: y_data[-1],
        batch_size=32,
        allow_partial_final_batch=True,
        shuffle=True,
        win_augment_callback=None,
    ):
        # Force in/out cols to be a list if just a single string
        if not isinstance(in_cols, list):
            in_cols = [in_cols]

        if not isinstance(out_cols, list):
            out_cols = [out_cols]

        # Add check if x_shape is a tuple

        self.win_inc = win_inc
        self.win_len = win_len
        self.in_cols = in_cols
        self.out_cols = out_cols
        self.weight_col = weight_col
        self.weight_callback = weight_callback
        self.out_callback = out_callback
        self.batch_size = batch_size
        self.shuffle = shuffle
        self.win_augment_callback = win_augment_callback

        # Get metadata about where files start and how long they are to make look up
        # quicker later
        # Assume that dataframe is already lex-sorted
        if len(df.index.names) <= 1:
            raise AttributeError(
                "DataFrame must use a multi-index, where the combination of "
                + "levels above the lowest represents a unique file."
            )
        else:
            self.file_indexes = sorted(set([x[:-1] for x in df.index]))
            

        self.org_index = df.index
        self.wins_per_file = np.zeros((len(self.file_indexes)))
        self.file_raw_start_idxs = np.zeros((len(self.file_indexes)))
        cur_raw_start = 0
        for i, file_index in enumerate(self.file_indexes):
            file_length = df.loc[(file_index)].shape[0]
            if file_length < win_len:
                self.wins_per_file[i] = 0

            else:
                self.wins_per_file[i] = int(
                    1 + np.floor((file_length - win_len) / win_inc)
                )

            self.file_raw_start_idxs[i] = cur_raw_start
            cur_raw_start += file_length

        self.nb_windows = int(np.sum(self.wins_per_file))

        # Which window number a file starts at
        self.file_win_start_nb = np.array(
            [0]
            + [sum(self.wins_per_file[:i]) for i in range(1, len(self.wins_per_file))]
        )

        # Split out data we want (may need a copy)
        self.x = df[in_cols].to_numpy()
        self.y = df[out_cols].to_numpy()
        
        # If using weighting check that the output is 1 weight per window
        if weight_col is not None:
            self.w = df[weight_col].to_numpy()
            if weight_callback(self.w[:win_len, ...]).shape != ():
                raise AttributeError(
                    "The weight_callback must return a single weight for each window"
                )

        # Determine desired x data shape
        if x_shape is None:
            self.x_shape = (win_len * len(in_cols),)

        else:
            self.x_shape = x_shape

        # Calculate desired y shape based on out_callback return
        self.y_shape = out_callback(self.y[:win_len, ...]).shape

        # Have a constant order to make sure each data point gets used
        self.window_order = np.array(list(range(self.nb_windows)))

        # Work out number of batches need
        self.allow_partial_final_batch = allow_partial_final_batch
        if allow_partial_final_batch:
            self.batches_per_epoch = int(np.ceil(self.nb_windows / self.batch_size))
            self.ignored_windows = 0

        else:
            self.batches_per_epoch = int(np.floor(self.nb_windows / self.batch_size))
            self.ignored_windows = self.nb_windows % self.batch_size

        # Run epoch end to handle anything we would normally do between runs such as
        # shuffle the data
        self.on_epoch_end()

    # Number of batches per epoch (lose last few samples if not divisible)
    def __len__(self):
        return self.batches_per_epoch

    # Make one batch
    def __getitem__(self, idx):
        cur_batch_size = self.batch_size
        if self.allow_partial_final_batch and idx == self.batches_per_epoch - 1:
            cur_batch_size = self.nb_windows - (idx * self.batch_size)

        x_batch = np.zeros((cur_batch_size,) + self.x_shape)
        y_batch = np.zeros((cur_batch_size,) + self.y_shape)
        
        if self.weight_col is not None:
            w_batch = np.zeros((cur_batch_size,))

        # Batches go in order to make it conceptually simpler
        cur_order_index = idx * self.batch_size
        for i in range(cur_batch_size):
            win_nb = self.window_order[cur_order_index + i]
            win_start_idx, win_end_idx = self._get_window_start_and_end(win_nb)

            # Add to batch
            x_batch[i, ...] = self.x[win_start_idx:win_end_idx, ...].reshape(
                self.x_shape
            )
            if self.win_augment_callback:
                x_batch[i, ...] = self.win_augment_callback(x_batch[i, ...])

            y_batch[i, ...] = self.out_callback(self.y[win_start_idx:win_end_idx, ...])
            
            if self.weight_col is not None:
                w_batch[i, ...] = self.weight_callback(self.w[win_start_idx:win_end_idx, ...])

        if self.weight_col is not None:
            return x_batch, y_batch, w_batch
        
        else:
            return (
                x_batch,
                y_batch,
            )  # Don't return none for class/sample weights as this breaks a lot of stuff

    # Update indices at the end of each epoch
    def on_epoch_end(self):
        if self.shuffle == True:
            np.random.shuffle(self.window_order)

    def get_windowed_dataset(self, return_x=True, return_w=False):
        """Get the whole dataset as windows (preserve shuffling and batch augments).
        
        Data returned in the order `x_all, y_all, w_all` with unreturned arrays being
        omitted
        
        - `return_x`: if True then return windowed x data; this may be large
        - `return_w`: if True then return the weighting information
        """
        if self.allow_partial_final_batch:
            nb_ret_windows = self.nb_windows

        else:
            nb_ret_windows = self.__len__() * self.batch_size

        if return_x:
            x_all = np.zeros((nb_ret_windows,) + self.x_shape)
            
        if return_w:
            w_all = np.zeros((nb_ret_windows,))

        y_all = np.zeros((nb_ret_windows,) + self.y_shape)

        for i in range(self.__len__()):
            if self.weight_col is not None:
                x_batch, y_batch, w_batch = self.__getitem__(i)
                
            else:
                x_batch, y_batch = self.__getitem__(i)

            nb_in_batch = x_batch.shape[0]
            start_idx = i * self.batch_size
            end_idx = (i + 1) * self.batch_size

            if return_x:
                x_all[start_idx:end_idx, ...] = x_batch
                
            if return_w:
                w_all[start_idx:end_idx, ...] = w_batch

            y_all[start_idx:end_idx, ...] = y_batch
        
        
        if return_x and return_w:
            return x_all, y_all, w_all
        
        elif return_x:
            return x_all, y_all
        
        elif return_w:
            return y_all, w_all

        else:
            return y_all

    def get_y_true(self):
        """Get the y data as windows (preserve shuffling).
        """
        return self.get_windowed_dataset(return_x=False)

    def get_file_divs(self):
        """Get the two synced lists that show where files are in the data.

        Returns:

        * `file_indexes`: list of file indices in the data (`N-1` levels of
          multi-index)
        * `file_divs`: 2D array of the start index (column 0) and end index
          (column 1) for each file in the dataset

        E.g. use `x_all[file_divs[0, 0]:file_divs[0, 1], :]` to get all the
        windows in file `file_indexes[0]`
        """

        if self.shuffle == True:
            raise RuntimeError(
                "Generator set to shuffle, this makes file divisions unhelpful"
            )

        file_divs = np.zeros((len(self.file_indexes), 2), dtype=np.int64)

        last_file_start = 0
        for i in range(len(self.file_indexes)):
            file_divs[i, 0] = int(last_file_start)
            file_divs[i, 1] = int(last_file_start + self.wins_per_file[i])

            last_file_start = file_divs[i][1]

        return self.file_indexes, file_divs

    def get_window_multi_indexs(self):
        """Get list of the original indices in the DataFrame for each window.

        Assumes only one returned multi-index per window based on out_callback
        """
        win_start_idx, win_end_idx = self._get_window_start_and_end(0)

        multi_index = np.zeros((self.nb_windows,), dtype=tuple)

        for i, win_nb in enumerate(self.window_order):
            win_start_idx, win_end_idx = self._get_window_start_and_end(win_nb)

            multi_index[i] = tuple(
                self.out_callback(self.org_index[win_start_idx:win_end_idx])
            )

        return pd.MultiIndex.from_tuples(multi_index, names=self.org_index.names)

    def get_window_idxs(self):
        """Get a 2D array of all window locations in the data.

        Returns a 2D list where the first column is the start index, the second
        is the exclusive end index.
        """
        windows = np.zeros((self.nb_windows, 2), dtype=np.int64)

        for i, win_nb in enumerate(self.window_order):
            windows[i, 0], windows[i, 1] = self._get_window_start_and_end(win_nb)

        return windows

    def register_augment_callback(self, new_callback):
        """Add a new augment callback (identical to if set in initialisation)
        """
        self.win_augment_callback = new_callback

    def deregister_augment_callback(self):
        """Remove augment callback (if any)
        """
        self.win_augment_callback = None

    def _get_window_start_and_end(self, window_nb):
        file_number = np.max(np.argmax(self.file_win_start_nb > window_nb) - 1, 0)
        file_raw_start_index = self.file_raw_start_idxs[file_number]

        win_start_idx = int(
            file_raw_start_index
            + self.win_inc * (window_nb - self.file_win_start_nb[file_number])
        )
        win_end_idx = int(win_start_idx + self.win_len)

        return win_start_idx, win_end_idx


def make_train_val_test_generators(
    train_data,
    val_data,
    test_data,
    **kwargs,
):  # pragma: no cover
    """Utility function to make train / validation / test generators in one go.
    
    Forwards all keywords to the generators but sets the train set generator to
    shuffling with no partial final batch and val / test generators to no shuffling
    and a partial final batch.
    """
    train_win_gen = MultiIndexDFWindowGenerator(
        train_data,
        allow_partial_final_batch=False,
        shuffle=True,
        **kwargs,
    )

    val_win_gen = MultiIndexDFWindowGenerator(
        val_data,
        allow_partial_final_batch=True,
        shuffle=False,
        **kwargs,
    )

    test_win_gen = MultiIndexDFWindowGenerator(
        test_data,
        allow_partial_final_batch=True,
        shuffle=False,
        **kwargs,
    )

    return train_win_gen, val_win_gen, test_win_gen


class MultiIndexDFCoherentWindowGenerator(MultiIndexDFWindowGenerator):
    """Extension of MultiIndexDFWindowGenerator to allow coherent outputs.

    Sometimes it is useful to include a loss term for temporal coherence. That
    is to include a loss term for the activations of subsequent windows being
    significantly different. The idea is that since nearby windows include a lot
    of overlapping information the network should be forced to encode that.

    Same arguments as MultiIndexDFWindowGenerator with the addition of:

    * `nb_coherent`: how many coherent steps to return (that is how many
      sequential windows) for each batch
        - Served as a list of length `nb_coherent` each element containing a
          synced batch
        - `[batch_t1, batch_t2]`
    * `y_coherent=True`: if `True` return `y` in the same way as `x`, if `False`
      then only include the batch for `t1` as MultiIndexDFWindowGenerator
    """

    def __init__(self, *args, nb_coherent=2, y_coherent=True, **kwargs):
        self.nb_coherent = nb_coherent
        self.y_coherent = y_coherent

        super(self.__class__, self).__init__(*args, **kwargs)

        # Redo indexing and windowing
        for i, file_index in enumerate(self.file_indexes):
            file_length = args[0].loc[(file_index)].shape[0] - (
                self.win_inc * (self.nb_coherent - 1)
            )
            if file_length < self.win_len:
                self.wins_per_file[i] = 0

            else:
                self.wins_per_file[i] = int(
                    1 + np.floor((file_length - self.win_len) / self.win_inc)
                )

        self.nb_windows = int(np.sum(self.wins_per_file))
        self.window_order = np.array(list(range(self.nb_windows)))

        self.file_win_start_nb = np.array(
            [0]
            + [sum(self.wins_per_file[:i]) for i in range(1, len(self.wins_per_file))]
        )

        # Recalculate number of batches needed
        self.allow_partial_final_batch = kwargs["allow_partial_final_batch"]
        if self.allow_partial_final_batch:
            self.batches_per_epoch = int(np.ceil(self.nb_windows / self.batch_size))
            self.ignored_windows = 0

        else:
            self.batches_per_epoch = int(np.floor(self.nb_windows / self.batch_size))
            self.ignored_windows = self.nb_windows % self.batch_size

        # Run epoch end in case of shuffling etc
        self.on_epoch_end()

    def __getitem__(self, idx):
        cur_batch_size = self.batch_size
        if self.allow_partial_final_batch and idx == self.batches_per_epoch - 1:
            cur_batch_size = self.nb_windows - (idx * self.batch_size)

        x_batch = []
        y_batch = []

        # Batches go in order to make it conceptually simpler
        for i in range(self.nb_coherent):
            x_batch_cur = np.zeros((cur_batch_size,) + self.x_shape)
            y_batch_cur = np.zeros((cur_batch_size,) + self.y_shape)

            cur_order_index = idx * self.batch_size
            for j in range(cur_batch_size):
                win_nb = self.window_order[cur_order_index + j]
                win_start_idx, win_end_idx = self._get_window_start_and_end(win_nb)
                win_start_idx += self.win_inc * i
                win_end_idx += self.win_inc * i

                # Add to batch
                x_batch_cur[j, ...] = self.x[win_start_idx:win_end_idx, ...].reshape(
                    self.x_shape
                )

                y_batch_cur[j, ...] = self.out_callback(
                    self.y[win_start_idx:win_end_idx, ...]
                )

            x_batch.append(x_batch_cur)
            y_batch.append(y_batch_cur)

        if self.y_coherent:
            return x_batch, y_batch

        else:
            return x_batch, y_batch[0]  # TODO do this more efficiently
